#!/usr/bin/python


""""------------------------------------------------PACKAGES---------------------------------------------------------"""
from __future__ import print_function
from scipy import stats
from pylab import *
import csv

"""---------------------------------------------------------------------------------------------------------------------
---------------------------------------------------FUNCTIONS---------------------------------------------------------"""


def combine_all(ori_rrna_range, ori_dnaa_range, ori_go_range, dnaa_max_pos,
                ORI_gc, ORI_ta, p_ori_gc, p_ori_ta, p_ori_go, ter_rrna_range,
                ter_dif_range, ter_go_range, dif_max_pos,
                TER_gc, TER_ta, p_ter_gc, p_ter_ta, p_ter_go, alpha, csv_file, sequence_length,
                av_diff_gc, av_diff_ta, linear=False, motifs=False, skews=False, allmeas=False, geneori=False):

    final_ORI = final_TER = -1
    final_p_ori = final_p_ter = -1.0
    use_ori = use_ter = True
    comment_ori = ""
    comment_ter = ""

    if p_ori_gc == p_ori_ta == 1.0:
        '''check if the intervals är ok'''
        if p_ori_go <= alpha:
            ''' use ori_go '''
            if len(ori_dnaa_range)  != 0:
                dist = list()
                for pos in ori_go_range:
                    dist.append(abs(pos-dnaa_max_pos))

                final_ORI = ori_go_range[dist.index(min(dist))]
                final_p_ori = p_ori_go
            else:
                final_ORI = round(mean([ori_go_range[0], ori_go_range[-1]]))
                final_p_ori = p_ori_go
        else:
            if len(ori_dnaa_range) != 0:
                '''use dnaa range'''
                if dnaa_max_pos in ori_dnaa_range:
                    final_ORI = dnaa_max_pos
                    final_p_ori = -1.0
                else:
                    final_ORI = round(mean([ori_dnaa_range[0], ori_dnaa_range[-1]]))
            else:
                if len(ori_rrna_range) != 0:
                    if ORI_gc in ori_rrna_range:
                        if ORI_ta in ori_rrna_range:
                            final_ORI = round(mean([ORI_gc, ORI_ta]))
                            stat, final_p_ori = stats.combine_pvalues([p_ori_gc, p_ori_ta], method='fisher',
                                                                      weights=None)
                        else:
                            final_ORI = ORI_gc
                            final_p_ori = p_ori_gc
                    elif ori_go_range in ori_rrna_range:
                        final_ORI = round(mean([ori_go_range[0], ori_go_range[-1]]))
                        final_p_ori = p_ori_go
                    else:
                        '''use rrna range'''
                        final_ORI = round(mean([ori_rrna_range[0], ori_rrna_range[-1]]))
                        final_p_ori = -1.0
                else:
                    '''cannot predict'''
                    final_ORI = -1
                    final_p_ori = -1.0
                    use_ori = False

    elif p_ori_gc != 1.0 == p_ori_ta:
        if ORI_gc in ori_go_range and p_ori_go <= alpha:
            if len(ori_dnaa_range) != 0:
                dist = list()
                for pos in ori_go_range:
                    dist.append(abs(pos-dnaa_max_pos))

                ORI_go = ori_go_range[dist.index(min(dist))]
                final_ORI = round(mean([ORI_go, ORI_gc]))
                stat, final_p_ori = stats.combine_pvalues([p_ori_gc, p_ori_go], method='fisher', weights=None)
            else:
                final_ORI = ORI_gc
                final_p_ori = p_ori_gc
        elif ORI_gc not in ori_go_range and p_ori_go <= alpha:
            if p_ori_gc <= alpha:
                if len(ori_dnaa_range) != 0:
                    dist = list()
                    for pos in ori_go_range:
                        dist.append(abs(pos - dnaa_max_pos))

                    ORI_go = ori_go_range[dist.index(min(dist))]
                    final_ORI = round(mean([ORI_go, ORI_gc]))
                    stat, final_p_ori = stats.combine_pvalues([p_ori_gc, p_ori_go], method='fisher', weights=None)
                else:
                    final_ORI = round(mean([ORI_gc, ori_go_range[0], ori_go_range[-1]]))
                    stat, final_p_ori = stats.combine_pvalues([p_ori_gc, p_ori_go], method='fisher', weights=None)
            else:
                if len(ori_dnaa_range) != 0:
                    dist = list()
                    for pos in ori_go_range:
                        dist.append(abs(pos - dnaa_max_pos))

                    ORI_go = ori_go_range[dist.index(min(dist))]
                    final_ORI = ORI_go
                    final_p_ori = p_ori_go
                else:
                    final_ORI = round(mean([ori_go_range[0], ori_go_range[-1]]))
                    final_p_ori = p_ori_go
        else:
            final_ORI = ORI_gc
            final_p_ori = p_ori_gc

    elif p_ori_gc == 1.0 != p_ori_ta:
        '''use ta'''
        if ORI_ta in ori_go_range and p_ori_go <= alpha:
            if len(ori_dnaa_range) != 0:
                dist = list()
                for pos in ori_go_range:
                    dist.append(abs(pos-dnaa_max_pos))

                ORI_go = ori_go_range[dist.index(min(dist))]
                final_ORI = round(mean([ORI_go, ORI_ta]))
                stat, final_p_ori = stats.combine_pvalues([p_ori_ta, p_ori_go], method='fisher', weights=None)
            else:
                final_ORI = ORI_ta
                final_p_ori = p_ori_ta
        elif ORI_ta not in ori_go_range and p_ori_go <= alpha:
            if p_ori_ta <= alpha:
                if len(ori_dnaa_range) != 0:
                    dist = list()
                    for pos in ori_go_range:
                        dist.append(abs(pos - dnaa_max_pos))

                    ORI_go = ori_go_range[dist.index(min(dist))]
                    final_ORI = round(mean([ORI_go, ORI_ta]))
                    stat, final_p_ori = stats.combine_pvalues([p_ori_ta, p_ori_go], method='fisher', weights=None)
                else:
                    final_ORI = round(mean([ORI_ta, ori_go_range[0], ori_go_range[-1]]))
                    stat, final_p_ori = stats.combine_pvalues([p_ori_ta, p_ori_go], method='fisher', weights=None)
            else:
                if len(ori_dnaa_range) != 0:
                    dist = list()
                    for pos in ori_go_range:
                        dist.append(abs(pos - dnaa_max_pos))

                    ORI_go = ori_go_range[dist.index(min(dist))]
                    final_ORI = ORI_go
                    final_p_ori = p_ori_go
                else:
                    final_ORI = round(mean([ori_go_range[0], ori_go_range[-1]]))
                    final_p_ori = p_ori_go
        else:
            final_ORI = ORI_ta
            final_p_ori = p_ori_ta
    else:
        '''cannot predict'''
        final_ORI = -1
        final_p_ori = -1.0
        use_ori = False

    '''TER'''
    if p_ter_gc == p_ter_ta == 1.0:
        '''Check if go pred significant '''
        if p_ter_go <= alpha:
            ''' Use'''
            if len(ter_dif_range) != 0:
                dist = list()
                for pos in ter_go_range:
                    dist.append(abs(pos-dif_max_pos))

                final_TER = ter_go_range[dist.index(min(dist))]
                final_p_ter = p_ter_go
            else:
                final_TER = round(mean([ter_go_range[0], ter_go_range[-1]]))
        else:
            '''cIf'''
            if len(ter_dif_range) != 0:
                if dif_max_pos in ter_dif_range:
                    final_TER = dif_max_pos
                    final_p_ter = -1.0
                else:
                    final_TER = round(mean([ter_dif_range[0], ter_dif_range[-1]]))
                    final_p_ter = -1.0
            else:
                if len(ter_rrna_range) != 0:
                    if TER_gc in ter_rrna_range:
                        if TER_ta in ter_rrna_range:
                            final_TER = round(mean([TER_gc, TER_ta]))
                            stat, final_p_ter = stats.combine_pvalues([p_ter_gc, p_ter_ta], method='fisher',
                                                                      weights=None)
                        else:
                            final_TER = TER_gc
                            final_p_ter = p_ter_gc
                    elif ter_go_range in ter_rrna_range:
                        final_TER = round(mean([TER_go[0], TER_go[1]]))
                        final_p_ter = p_ter_go
                    else:
                        '''use rrna range'''
                        final_TER = round(mean([ter_rrna_range[0], ter_rrna_range[-1]]))
                        final_p_ter = -1.0
                else:
                    '''cannot predict'''
                    final_TER = -1
                    final_p_ter = -1.0
                    use_ter = False

    elif p_ter_gc != 1.0 == p_ter_ta:
        if TER_gc in ter_go_range and p_ter_go <= alpha:
            if len(ter_dif_range) != 0:
                dist = list()
                for pos in ter_go_range:
                    dist.append(abs(pos-dif_max_pos))

                TER_go = ter_go_range[dist.index(min(dist))]
                final_TER = round(mean([TER_go, TER_gc]))
                stat, final_p_ter = stats.combine_pvalues([p_ter_gc, p_ter_go], method='fisher', weights=None)
            else:
                final_TER = TER_gc
                final_p_ter = p_ter_gc

        elif TER_gc not in ter_go_range and p_ter_go <= alpha:
            if p_ter_gc <= alpha:
                if len(ter_dif_range) != 0:
                    dist = list()
                    for pos in ter_go_range:
                        dist.append(abs(pos - dif_max_pos))

                    TER_go = ter_go_range[dist.index(min(dist))]
                    final_TER = round(mean([TER_go, TER_gc]))
                    stat, final_p_ter = stats.combine_pvalues([p_ter_gc, p_ter_go], method='fisher', weights=None)
                else:
                    final_TER = round(mean([TER_gc, ter_go_range[0], ter_go_range[-1]]))
                    stat, final_p_ter = stats.combine_pvalues([p_ter_gc, p_ter_go], method='fisher', weights=None)
            else:
                if len(ter_dif_range) != 0:
                    dist = list()
                    for pos in ter_go_range:
                        dist.append(abs(pos - dif_max_pos))

                    TER_go = ter_go_range[dist.index(min(dist))]
                    final_TER = TER_go
                    final_p_ter = p_ter_go
                else:
                    final_TER = round(mean([ter_go_range[0], ter_go_range[-1]]))
                    final_p_ter = p_ter_go
        else:
            final_TER = TER_gc
            final_p_ter = p_ter_gc

    elif p_ter_gc == 1.0 != p_ter_ta:
        '''use ta'''
        if TER_ta in ter_go_range and p_ter_go <= alpha:
            if len(ter_dif_range) != 0:
                dist = list()
                for pos in ter_go_range:
                    dist.append(abs(pos-dif_max_pos))

                TER_go = ter_go_range[dist.index(min(dist))]
                final_TER = round(mean([TER_go, TER_ta]))
                stat, final_p_ter = stats.combine_pvalues([p_ter_ta, p_ter_go], method='fisher', weights=None)
            else:
                final_TER = TER_ta
                final_p_ter = p_ter_ta
        elif TER_ta not in ter_go_range and p_ter_go <= alpha:
            if p_ter_ta <= alpha:
                if len(ter_dif_range) != 0:
                    dist = list()
                    for pos in ter_go_range:
                        dist.append(abs(pos - dif_max_pos))

                    TER_go = ter_go_range[dist.index(min(dist))]
                    final_TER = round(mean([TER_go, TER_ta]))
                    stat, final_p_ter = stats.combine_pvalues([p_ter_ta, p_ter_go], method='fisher', weights=None)
                else:
                    final_TER = round(mean([TER_ta, ter_go_range[0], ter_go_range[-1]]))
                    stat, final_p_ter = stats.combine_pvalues([p_ter_ta, p_ter_go], method='fisher', weights=None)
            else:
                if len(ter_dif_range) != 0:
                    dist = list()
                    for pos in ter_go_range:
                        dist.append(abs(pos - dif_max_pos))

                    TER_go = ter_go_range[dist.index(min(dist))]
                    final_TER = TER_go
                    final_p_ter = p_ter_go
                else:
                    final_TER = round(mean([ter_go_range[0], ter_go_range[-1]]))
                    final_p_ter = p_ter_go
        else:
            final_TER = TER_ta
            final_p_ter = p_ter_ta
    else:
        '''cannot predict'''
        final_TER = -1
        final_p_ter = -1.0
        use_ter = False

    if not use_ori:
        comment_ori = "ORI prediction not possible. "
    if not use_ter:
        comment_ter = "TER prediction not possible."

    comment = ""
    if not use_ori and not use_ter:
        comment = "Predictions not possible."
    else:
        comment = comment_ori + comment_ter

    if use_ori:
        if final_p_ori > alpha:
            comment_ori = "NB! ORI p-value above significance level, \n" \
                          "do not trust result. "
    if use_ter:
        if final_p_ter > alpha:
            comment_ter = "NB! TER p-value above significance level, \n" \
                          "do not trust result. "

    if use_ori and use_ter and final_p_ori > alpha < final_p_ter:
        comment = "NB! Both p-values indicate insignificant result, \n" \
                  "do not trust result!"
    else:
        comment = comment_ori + comment_ter

    if final_p_ori <= alpha or final_p_ter <= alpha:
        if (av_diff_gc <= av_diff_ta and av_diff_gc > 0.05) or \
                (av_diff_ta < av_diff_gc and av_diff_ta > 0.05):
            comment = comment + "Large variation -- do not trust p-values."

    '''No p-values if only running motifs'''
    if motifs == True and not skews and not allmeas and not geneori:
        final_p_ori = final_p_ter = -1.0

    predict(int(final_ORI), int(final_TER), round(final_p_ori, 3), round(final_p_ter, 3), comment,
            csv_file, sequence_length, use_ori, use_ter, "", linear)


def predict(ORI, TER, p_ori, p_ter, comment, csv_file, sequence_length,
            use_ori=True, use_ter=True, pred="", linear=False):
    """
    Adds final prediction to csv file

    :param ORI:             Final prediction of ORI
    :param TER:             Final prediction of TER
    :param p_ori:           Final p-value for ORI prediction
    :param p_ter:           Final p-value for TER prediction
    :param comment:         Comment if needed
    :param csv_file:        Name of csv file
    :param sequence_length: Length of sequence studied
    :param use_ori:         Is the ORI prediction ok or not
    :param use_ter:         Is the TER prediction ok or not
    :param pred:            Both predictions, only ORI or only TER
    :param linear:          If the sequence is from a linear chromosome
    :return:                Saves to file
    """
    if linear:
        if TER != sequence_length - 1:
            TER_1 = TER
            TER_2 = TER + 1
        elif TER == sequence_length - 1:
            TER_1 = TER
            TER_2 = 0

        if use_ori and use_ter:
            if p_ori == p_ter == -1.0:
                with open(csv_file, mode='a') as results:
                    teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                    teritori_writer.writerow(["", ORI, "", TER_1, TER_2, "", comment])
            elif p_ori == -1.0 and p_ter != -1.0:
                with open(csv_file, mode='a') as results:
                    teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                    teritori_writer.writerow(["", ORI, "", TER_1, TER_2, p_ter, comment])
            elif p_ori != -1.0 and p_ter == -1.0:
                with open(csv_file, mode='a') as results:
                    teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                    teritori_writer.writerow(["", ORI, p_ori, TER_1, TER_2, "", comment])
            else:
                with open(csv_file, mode='a') as results:
                    teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                    teritori_writer.writerow(["", ORI, p_ori, TER_1, TER_2, p_ter, comment])
        elif use_ori and not use_ter:
            if p_ori == -1.0:
                if pred == "":
                    with open(csv_file, mode='a') as results:
                        teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                        teritori_writer.writerow(["", ORI, "", "---", "---", "---", comment])
                elif pred == "ori":
                    with open(csv_file, mode='a') as results:
                        teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                        teritori_writer.writerow(["", ORI, "", "", "", "", comment])
            else:
                if pred == "":
                    with open(csv_file, mode='a') as results:
                        teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                        teritori_writer.writerow(["", ORI, p_ori, "---", "---", "---", comment])
                elif pred == "ori":
                    with open(csv_file, mode='a') as results:
                        teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                        teritori_writer.writerow(["", ORI, p_ori, "", "", "", comment])
                else:
                    with open(csv_file, mode='a') as results:
                        teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                        teritori_writer.writerow(["", ORI, p_ori, "---", "---", "---", comment])
        elif not use_ori and use_ter:
            if p_ter == -1.0:
                if pred == "":
                    with open(csv_file, mode='a') as results:
                        teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                        teritori_writer.writerow(["", "---", "---", TER_1, TER_2, "", comment])
                elif pred == "ter":
                    with open(csv_file, mode='a') as results:
                        teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                        teritori_writer.writerow(["", "", "", TER_1, TER_2, "", comment])
            else:
                if pred == "":
                    with open(csv_file, mode='a') as results:
                        teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                        teritori_writer.writerow(["", "---", "---", TER_1, TER_2, p_ter, comment])
                elif pred == "ter":
                    with open(csv_file, mode='a') as results:
                        teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                        teritori_writer.writerow(["", "", "", TER_1, TER_2, p_ter, comment])
                else:
                    with open(csv_file, mode='a') as results:
                        teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                        teritori_writer.writerow(["", "---", "---", TER_1, TER_2, p_ter, comment])
        elif not use_ori and not use_ter:
            with open(csv_file, mode='a') as results:
                teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                teritori_writer.writerow(["", "---", "---", "---", "---", "---", "Predictions not possible"])
    elif not linear:
        if use_ori and use_ter:
            if p_ori == p_ter == -1.0:
                with open(csv_file, mode='a') as results:
                    teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                    teritori_writer.writerow(["", ORI, "", TER, "", comment])
            elif p_ori == -1.0 and p_ter != -1.0:
                with open(csv_file, mode='a') as results:
                    teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                    teritori_writer.writerow(["", ORI, "", TER, p_ter, comment])
            elif p_ori != -1.0 and p_ter == -1.0:
                with open(csv_file, mode='a') as results:
                    teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                    teritori_writer.writerow(["", ORI, p_ori, TER, "", comment])
            else:
                with open(csv_file, mode='a') as results:
                    teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                    teritori_writer.writerow(["", ORI, p_ori, TER, p_ter, comment])
        elif use_ori and not use_ter:
            if p_ori == -1.0:
                if pred == "":
                    with open(csv_file, mode='a') as results:
                        teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                        teritori_writer.writerow(["", ORI, "", "---", "---", comment])
                elif pred == "ori":
                    with open(csv_file, mode='a') as results:
                        teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                        teritori_writer.writerow(["", ORI, "", "", "", comment])
            else:
                if pred == "":
                    with open(csv_file, mode='a') as results:
                        teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                        teritori_writer.writerow(["", ORI, p_ori, "---", "---", comment])
                elif pred == "ori":
                    with open(csv_file, mode='a') as results:
                        teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                        teritori_writer.writerow(["", ORI, p_ori, "", "", comment])
                else:
                    with open(csv_file, mode='a') as results:
                        teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                        teritori_writer.writerow(["", ORI, p_ori, "---", "---", comment])
        elif not use_ori and use_ter:
            if p_ter == -1.0:
                if pred == "":
                    with open(csv_file, mode='a') as results:
                        teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                        teritori_writer.writerow(["", "---", "---", TER, "", comment])
                elif pred == "ter":
                    with open(csv_file, mode='a') as results:
                        teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                        teritori_writer.writerow(["", "", "", TER, "", comment])
            else:
                if pred == "":
                    with open(csv_file, mode='a') as results:
                        teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                        teritori_writer.writerow(["", "---", "---", TER, p_ter, comment])
                elif pred == "ter":
                    with open(csv_file, mode='a') as results:
                        teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                        teritori_writer.writerow(["", "", "", TER, p_ter, comment])
                else:
                    with open(csv_file, mode='a') as results:
                        teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                        teritori_writer.writerow(["", "---", "---", TER, p_ter, comment])
        elif not use_ori and not use_ter:
            with open(csv_file, mode='a') as results:
                teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                teritori_writer.writerow(["", "---", "---", "---", "---", "Predictions not possible"])


#!/usr/bin/python


"""----------------------------------------------PACKAGES------------------------------------------------------------"""
from __future__ import print_function
import timeit
import matplotlib.pyplot as plt
import numpy as np
import sys
import subprocess
import Bio
from random import choice
from scipy import stats
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
from sklearn.preprocessing import MinMaxScaler, minmax_scale


"""---------------------------------------------------------------------------------------------------------------------
--------------------------------------------------FUNCTIONS----------------------------------------------------------"""


def fill_positions_dict(genes_dictionary):
    """
    Fills a dictionary with the gene positions
    :param genes_dictionary:    Dictionary containing all genes
    :return:
    """
    positions_dictionary = {}

    for key in genes_dictionary:
        split = genes_dictionary[key].description.split(" # ")

        original_start = int(split[1])
        original_end = int(split[2])
        strand = int(split[3])
        length_gene = len(genes_dictionary[key].seq)

        positions_dictionary[key] = [original_start, original_end, length_gene, strand]

    return positions_dictionary


def calculate_overall_gc(genes_dictionary, positions_dictionary, sequence):
    """

    :param genes_dictionary:        Dictionary with all genes
    :param positions_dictionary:    Dictionary with all gene positions
    :param sequence:                FASTA sequence of the chromosome
    :return:                        G, C, T and A content of the chromosome
    """

    # Initiates count
    G_plus = G_minus = C_plus = C_minus = 0

    # Length of chromosome
    sequence_length = len(sequence)

    # Calculates the relative amount of the bases in the sequence
    G_content = round(((sequence.count('G') + sequence.count('g'))/sequence_length)*100)
    C_content = round(((sequence.count('C') + sequence.count('c'))/sequence_length)*100)
    T_content = round(((sequence.count('T') + sequence.count('t'))/sequence_length)*100)
    A_content = 100 - G_content - C_content - T_content

    return G_content, C_content, T_content, A_content


def prod_gc_skew(genes_dictionary, positions, window_size, sequence_length, sequence):
    """
    Produces a list of gc-skew and ta-skew values.

    :param genes_dictionary:    Dictionary with all genes from Prodigal
    :param positions:           Dictionary for positions in whole genome, in all_genes and in third_pos
    :param window_size:         Size of window for calculation
    :param sequence_length:     Length of chromosome
    :param sequence:            FASTA sequence of chromosome
    :return:                    Lists with gc- and ta-skews
    """

    # List for storing every third nt in all genes
    ev_third = ['u'] * sequence_length

    # Reverse of the sequence
    rev = list(sequence.seq)

    # Lists for saving gc- and ta-skews
    gc_skew_list = [0.0] * sequence_length
    ta_skew_list = [0.0] * sequence_length

    for key in genes_dictionary:
        '''If plus strand gene -- save every third nt to list'''
        if positions[key][3] == 1:
            for i in range(positions[key][0]+2, positions[key][1], 3):
                ev_third[i] = sequence.seq[i-1]

        '''If minus strand gene -- reverse list and save every third nt'''
        if positions[key][3] == -1:
            rev[positions[key][0]:positions[key][1]+1] = list(genes_dictionary[key].seq)
            for i in range(positions[key][0]+2, positions[key][1], 3):
                ev_third[i] = rev[i]

        '''Calculate the number of each nt per window for every 10th nt in the list'''
        for i in range(positions[key][0], positions[key][1]+1, 10):
            if i-(window_size//2) < 0:
                rest = abs(i - (window_size//2))
                a = ev_third[sequence_length - 1 - rest:i + (window_size // 2)].count('A') + \
                    ev_third[sequence_length - 1 - rest:i + (window_size // 2)].count('a') + \
                    ev_third[0:i + (window_size // 2)].count('A') + \
                    ev_third[0:i + (window_size // 2)].count('a')
                c = ev_third[sequence_length - 1 - rest:i + (window_size // 2)].count('C') + \
                    ev_third[sequence_length - 1 - rest:i + (window_size // 2)].count('c') + \
                    ev_third[0:i + (window_size // 2)].count('C') + \
                    ev_third[0:i + (window_size // 2)].count('c')
                g = ev_third[sequence_length - 1 - rest:i + (window_size // 2)].count('G') + \
                    ev_third[sequence_length - 1 - rest:i + (window_size // 2)].count('g') + \
                    ev_third[0:i + (window_size // 2)].count('G') + \
                    ev_third[0:i + (window_size // 2)].count('g')
                t = ev_third[sequence_length - 1 - rest:i + (window_size // 2)].count('T') + \
                    ev_third[sequence_length - 1 - rest:i + (window_size // 2)].count('t') + \
                    ev_third[0:i + (window_size // 2)].count('T') + \
                    ev_third[0:i + (window_size // 2)].count('t')
            elif i+(window_size//2) > sequence_length-1:
                rest = abs(sequence_length - (window_size // 2))
                a = ev_third[i-(window_size//2):sequence_length-1].count('A') + \
                    ev_third[i-(window_size//2):sequence_length-1].count('a') + \
                    ev_third[0:rest].count('A') + \
                    ev_third[0:rest].count('a')
                c = ev_third[i - (window_size // 2):sequence_length - 1].count('C') + \
                    ev_third[i - (window_size // 2):sequence_length - 1].count('c') + \
                    ev_third[0:rest].count('C') + \
                    ev_third[0:rest].count('c')
                g = ev_third[i - (window_size // 2):sequence_length - 1].count('G') + \
                    ev_third[i - (window_size // 2):sequence_length - 1].count('g') + \
                    ev_third[0:rest].count('G') + \
                    ev_third[0:rest].count('g')
                t = ev_third[i - (window_size // 2):sequence_length - 1].count('T') + \
                    ev_third[i - (window_size // 2):sequence_length - 1].count('t') + \
                    ev_third[0:rest].count('T') + \
                    ev_third[0:rest].count('t')
            else:
                a = ev_third[i-(window_size//2):i + (window_size//2)].count('A') + ev_third[i-(window_size//2):i + (window_size//2)].count('a')
                c = ev_third[i-(window_size//2):i + (window_size//2)].count('C') + ev_third[i-(window_size//2):i + (window_size//2)].count('c')
                g = ev_third[i-(window_size//2):i + (window_size//2)].count('G') + ev_third[i-(window_size//2):i + (window_size//2)].count('g')
                t = ev_third[i-(window_size//2):i + (window_size//2)].count('T') + ev_third[i-(window_size//2):i + (window_size//2)].count('t')

            '''If not division with zero -- calculate gc- and ta-skew and save to lists'''
            if (g + c) != 0.0:
                gc_skew_list[i] = positions[key][3] * ((g - c) / (g + c))

            if (t + a) != 0.0:
                ta_skew_list[i] = positions[key][3] * ((t - a) / (t + a))

    return gc_skew_list, ta_skew_list


def find_ter_ori(gc_skew_list, ta_skew_list):
    """
    Find positions of ORI and TER

    :param gc_skew_list:    List of gc-skews
    :param ta_skew_list:    List of ta-skews
    :return:                Positions of ORI and TER
    """

    # Producing cumulative gc skew
    gc_cum = list(np.cumsum(gc_skew_list))
    ta_cum = list(np.cumsum(ta_skew_list))

    # Finds minimum value in gc-skew
    ORI_gc = gc_cum.index(min(gc_cum))

    # Finds minimum value in ta-skew
    ORI_ta = ta_cum.index(min(ta_cum))

    # Find maximum value in gc-skew
    TER_gc = gc_cum.index(max(gc_cum))

    # Find maximum value in ta-skew
    TER_ta = ta_cum.index(max(ta_cum))

    return ORI_gc, ORI_ta, TER_gc, TER_ta, gc_cum, ta_cum


def find_actual_position(ORI_gc, ORI_ta, TER_gc, TER_ta, gc_cum, ta_cum, positions):
    """
    Find position of ORI and TER - assumed to be before gene where predicted position is in
    and after the gene where the predicted position is in, respectively

    :param ORI_gc:      Predicted position of ORI using GC-skew
    :param ORI_ta:      Predicted position of ORI using TA-skew
    :param TER_gc:      Predicted position of TER using GC-skew
    :param TER_ta:      Predicted position of TER using TA-skew
    :param gc_cum:      List of cumulative GC-skews
    :param ta_cum:      List of cumulative TA-skews
    :param positions:   List of gene positions
    :return:            Actual position for ORI and TER using both GC and TA-skew,
                        and the indices values in the list (used later for plotting)
    """

    # Initiating that none of the actual positions have been found
    ori_gc_found = ter_gc_found = ori_ta_found = ter_ta_found = False

    '''Loop through all genes, find the gene which contains the predicted position
    and set the new predicted position as the position before that gene (ORI) or after
    that gene (TER). Once all positions have been found, loop breaks.'''
    for gene in positions:
        if not ori_gc_found:
            if positions[gene][0] <= ORI_gc <= positions[gene][1]:
                ORI_gc = positions[gene][0]
                ori_gc_found = True
        if not ter_gc_found:
            if positions[gene][0] <= TER_gc <= positions[gene][1]:
                TER_gc = positions[gene][1]
                ter_gc_found = True
        if not ori_ta_found:
            if positions[gene][0] <= ORI_ta <= positions[gene][1]:
                ORI_ta = positions[gene][0]
                ori_ta_found = True
        if not ter_ta_found:
            if positions[gene][0] <= TER_ta <= positions[gene][1]:
                TER_ta = positions[gene][1]
                ter_ta_found = True
        if ori_gc_found and ter_gc_found and ori_ta_found and ter_ta_found:
            break

    ORI_gc = ORI_gc - 1
    ORI_ta = ORI_ta - 1
    TER_gc = TER_gc - 1
    TER_ta = TER_ta - 1
    # Finds the values for each index
    ORI_gc_val = gc_cum[ORI_gc]
    ORI_ta_val = ta_cum[ORI_ta]
    TER_gc_val = gc_cum[TER_gc]
    TER_ta_val = ta_cum[TER_ta]

    return ORI_gc, ORI_gc_val, TER_gc, TER_gc_val, ORI_ta, ORI_ta_val, TER_ta, TER_ta_val


def av_slope(normalized, first, second):
    """

    :param normalized:  List of normalized data (not necessary to be normalized)
    :param first:       First position - ORI or TER
    :param second:      Second position - ORI or TER
    :return:            Returns the slope
    """
    dy = normalized[second-1] - normalized[first]
    dx = second - first
    return dy / dx


def slopes(gc_normalized, ta_normalized, ORI_gc, TER_gc, ORI_ta, TER_ta, size=100000):
    """
    Calculates slopes before and after ORI and TER

    :param gc_normalized:   Normalized list of gc-skew values
    :param ta_normalized:   Normalized list of ta-skew values
    :param ORI_gc:          Predicted ORI position by gc-skew
    :param TER_gc:          Predicted TER position by gc-skew
    :param ORI_ta:          Predicted ORI position by ta-skew
    :param TER_ta:          Predicted TER position by ta-skew
    :param size:            Size of window to look at
    :return:                Slopes before and after ORI and TER for
                            GC and TA skew lists.
    """
    len_gc_normalized = len(gc_normalized)
    len_ta_normalized = len(ta_normalized)

    # Average overall slope
    average_slope_gc = av_slope(gc_normalized, 0, len_gc_normalized)
    average_slope_ta = av_slope(ta_normalized, 0, len_ta_normalized)

    # Average slope between ori and ter
    if ORI_gc <= TER_gc:
        av_ori_ter_gc = (gc_normalized[ORI_gc] - gc_normalized[TER_gc]) / (ORI_gc - TER_gc)
    elif ORI_gc > TER_gc:
        av_ori_ter_gc = (gc_normalized[TER_gc] - gc_normalized[ORI_gc]) / (TER_gc - ORI_gc)

    if ORI_ta <= TER_ta:
        av_ori_ter_ta = av_slope(ta_normalized, ORI_ta, TER_ta)
    elif ORI_ta > TER_ta:
        av_ori_ter_ta = av_slope(ta_normalized, TER_ta, ORI_ta)

    # Average slope before and after ori and ter for gc and ta skew
    if size < ORI_gc:
        av_gc_bf_ori = av_slope(gc_normalized, ORI_gc - size, ORI_gc)
    else:
        av_gc_bf_ori = av_slope(gc_normalized, 0, ORI_gc)

    # --------
    if ORI_gc < len_gc_normalized-(size+1):
        av_gc_af_ori = av_slope(gc_normalized, ORI_gc, ORI_gc + size)
    else:
        av_gc_af_ori = av_slope(gc_normalized, ORI_gc, len_gc_normalized)

    # --------
    if size < TER_gc:
        av_gc_bf_ter = av_slope(gc_normalized, TER_gc - size, TER_gc)
    else:
        av_gc_bf_ter = av_slope(gc_normalized, 0, TER_gc)

    # --------
    if TER_gc < len_gc_normalized - (size+1):
        av_gc_af_ter = av_slope(gc_normalized, TER_gc, TER_gc + size)
    else:
        av_gc_af_ter = av_slope(gc_normalized, TER_gc, len_gc_normalized)

    # --------
    if size < ORI_ta:
        av_ta_bf_ori = av_slope(ta_normalized, ORI_ta - size, ORI_ta)
    else:
        av_ta_bf_ori = av_slope(ta_normalized, 0, ORI_ta)

    # --------
    if ORI_ta < len_ta_normalized-(size+1):
        av_ta_af_ori = av_slope(ta_normalized, ORI_ta, ORI_ta + size)
    else:
        av_ta_af_ori = av_slope(ta_normalized, ORI_ta, len_ta_normalized)

    # --------
    if size < TER_ta:
        av_ta_bf_ter = av_slope(ta_normalized, TER_ta - size, TER_ta)
    else:
        av_ta_bf_ter = av_slope(ta_normalized, 0, TER_ta)

    # --------
    if TER_ta < len_ta_normalized - (size+1):
        av_ta_af_ter = av_slope(ta_normalized, TER_ta, TER_ta + size)
    else:
        av_ta_af_ter = av_slope(ta_normalized, TER_ta, len_ta_normalized)

    return av_gc_bf_ori, av_gc_af_ori, av_gc_bf_ter, av_gc_af_ter, av_ta_bf_ori, \
        av_ta_af_ori, av_ta_bf_ter, av_ta_af_ter, average_slope_gc, average_slope_ta, \
        av_ori_ter_gc, av_ori_ter_ta


def perfect_plot_norm(first, second, real_normalized):
    """
    Produces datasets with perfect, normalized data - Linear between start and the
    predicted positions, between the predicted positions, and the end of the chromosome

    :param first:           Position with lowest index
    :param second:          Position with highest index
    :param real_normalized: Normalized data
    :return:                Linear data set and the x-values
    """

    # Length of the normalized data set
    rn_length = len(real_normalized)

    # Defines a set of linear data points between start of sequence and first position
    start_first = real_normalized[0:first]
    x_start_first = list(np.arange(0, first, 1))
    slope, intercept = np.polyfit(x_start_first, start_first, 1)
    start_first_vals = [slope * i + intercept for i in x_start_first]

    # Defines a set of linear data points between first and second position
    first_second = real_normalized[first:second]
    x_first_second = list(np.arange(first, second, 1))
    slope, intercept = np.polyfit(x_first_second, first_second, 1)
    first_second_vals = [slope * i + intercept for i in x_first_second]

    # Defines a set of linear data points between second position and end of sequence
    second_end = real_normalized[second:rn_length]
    x_second_end = list(np.arange(second, rn_length, 1))
    slope, intercept = np.polyfit(x_second_end, second_end, 1)
    second_end_vals = [slope * i + intercept for i in x_second_end]

    # Puts the above datasets into two continuous ones
    perfect = start_first_vals + first_second_vals + second_end_vals
    x_perfect = np.arange(0, rn_length, 1)

    return perfect, x_perfect


def diff_perf_real(perfect_gc, real_gc_norm, perfect_ta, real_ta_norm):
    len_perfect_gc = len(perfect_gc)
    len_perfect_ta = len(perfect_ta)
    av_diff_gc = av_diff_ta = 0

    if len(real_gc_norm) == len_perfect_gc and len(real_ta_norm) == len_perfect_ta:
        sum_diff_gc = 0
        sum_diff_ta = 0
        for i in range(len_perfect_gc):
            sum_diff_gc = sum_diff_gc + abs(perfect_gc[i]-real_gc_norm[i])
        for i in range(len_perfect_ta):
            sum_diff_ta = sum_diff_ta + abs(perfect_ta[i]-real_ta_norm[i])
        av_diff_gc = sum_diff_gc/len_perfect_gc
        av_diff_ta = sum_diff_ta/len_perfect_ta
    return av_diff_gc, av_diff_ta


def gc_bootstrap(sequence_length, aroundter_gc, aroundori_gc,
                 aroundter_ta, aroundori_ta, window_size, G_content, C_content,
                 T_content, A_content, ori_break_gc, ter_break_gc, ori_break_ta, ter_break_ta,
                 ORI_gc, TER_gc, log_file, ori_rrna_range = list(), ter_rrna_range = list()):
    """
    Bootstrap function, produces randomly generated sequences with same relative amount of
    the different bases.

    :param sequence_length: Length of chromosome
    :param aroundter_gc:    List containing indexes around TER predicted by GC-skew
    :param aroundori_gc:    List containing indexes around ORI predicted by GC-skew
    :param aroundter_ta:    List containing indexes around TER predicted by TA-skew
    :param aroundori_ta:    List containing indexes around ORI predicted by TA-skew
    :param window_size:     Window size
    :param G_content:       Relative amount of G in sequence
    :param C_content:       Relative amount of C in sequence
    :param T_content:       Relative amount of T in sequence
    :param A_content:       Relative amount of A in sequence
    :param ori_break_gc:    Position of break in indices in aroundori_gc
    :param ter_break_gc:    Position of break in indices in aroundter_gc
    :param ori_break_ta:    Position of break in indices in aroundori_ta
    :param ter_break_ta:    Position of break in indices in aroundter_ta
    :param ORI_rrnas:       Predicted interval of ORI from rRNAs
    :param TER_rrnas:       Predicted interval of TER from rRNAs
    :param ORI_gc:          Predicted ORI by GC-skew
    :param TER_gc:          Predicted TER by GC-skew
    :param ORI_ta:          Predicted ORI by TA-skew
    :param TER_ta:          Predicted TER by TA-skew
    :return:                p-values for ORI and TER from GC- and TA- skew
    """

    # Number of iterations
    N = 1000

    # List for storing indices in rRNA range
    rrnas_ori = list()
    rrnas_ter = list()

    # Initiates p-values
    p_ter_gc = p_ori_gc = p_ter_ta = p_ori_ta = 0.0

    # All predictions and p-values are set to be used
    use_gc_ter = True
    use_gc_ori = True
    use_ta_ter = True
    use_ta_ori = True

    # Calculates maximum expected difference between ORI and TER
    max_perc_genome = round(sequence_length * 0.70)
    min_perc_genome = int(sequence_length * 0.30)

    '''If distance between the predicted positions is greater than the expected, 
    the p-values are set to 1.0'''
    if not min_perc_genome <= abs(ORI_gc - TER_gc) <= max_perc_genome:
        p_ter_gc = p_ori_gc = 1.0
        use_gc_ter = use_gc_ori = False
    # if not min_perc_genome <= abs(ORI_ta - TER_ta) <= max_perc_genome:
    #     p_ter_ta = p_ori_ta = 1.0
    #     use_ta_ter = use_ta_ori = False

    '''If the predicted positions are not in the rRNA prediction range, 
    the p-values are set to 1.0'''
    if len(ori_rrna_range) != 0:
        if ORI_gc not in ori_rrna_range:
            p_ori_gc = 1.0
            use_gc_ori = False
        # if ORI_ta not in ori_rrna_range:
        #     p_ori_ta = 1.0
        #     use_ta_ori = False
    if len(ter_rrna_range) != 0:
        if TER_gc not in ter_rrna_range:
            p_ter_gc = 1.0
            use_gc_ter = False
        # if TER_ta not in ter_rrna_range:
        #     p_ter_ta = 1.0
        #     use_ta_ter = False

    '''If none of the predictions are ok, returns the p-values, otherwise
    runs bootstrap'''
    if not use_gc_ter and not use_gc_ori:
        return p_ter_gc, p_ori_gc
    else:
        # Specifies which bases can be chosen
        bases = ['A', 'T', 'C', 'G']

        # Produces the weights depending on the original sequence composition
        weights = [A_content/100, T_content/100, C_content/100, G_content/100]

        with open(log_file, "a") as logs:
            logs.write("%s           %s" % ("", ".."))

        for i in range(0, N):
            if i % 100 == 0:
                with open(log_file, "a") as logs:
                    logs.write("{}..".format(i))

            # Initiates list for saving every third nt
            new_third = ['u'] * sequence_length

            # Randomly generates a new sequence
            newsample = np.random.choice(bases, sequence_length, p=weights, replace=True)

            # Saves every third nt
            for k in range(2, sequence_length, 3):
                new_third[k] = newsample[k]

            # Initiates list for storing skew values
            newsample_gc = [0.0] * sequence_length
            newsample_ta = [0.0] * sequence_length

            '''Counts the number of the different nt in the windows'''
            for j in range(0, sequence_length, window_size):
                if j-(window_size//2) < 0:
                    rest = abs(j - (window_size//2))
                    a = new_third[sequence_length - 1 - rest:j + (window_size // 2)].count('A') + \
                        new_third[sequence_length - 1 - rest:j + (window_size // 2)].count('a') + \
                        new_third[0:j + (window_size // 2)].count('A') + \
                        new_third[0:j + (window_size // 2)].count('a')
                    c = new_third[sequence_length - 1 - rest:j + (window_size // 2)].count('C') + \
                        new_third[sequence_length - 1 - rest:j + (window_size // 2)].count('c') + \
                        new_third[0:j + (window_size // 2)].count('C') + \
                        new_third[0:j + (window_size // 2)].count('c')
                    g = new_third[sequence_length - 1 - rest:j + (window_size // 2)].count('G') + \
                        new_third[sequence_length - 1 - rest:j + (window_size // 2)].count('g') + \
                        new_third[0:j + (window_size // 2)].count('G') + \
                        new_third[0:j + (window_size // 2)].count('g')
                    t = new_third[sequence_length - 1 - rest:j + (window_size // 2)].count('T') + \
                        new_third[sequence_length - 1 - rest:j + (window_size // 2)].count('t') + \
                        new_third[0:j + (window_size // 2)].count('T') + \
                        new_third[0:j + (window_size // 2)].count('t')
                elif j+(window_size//2) > sequence_length-1:
                    rest = abs(sequence_length - (window_size // 2))
                    a = new_third[j-(window_size//2):sequence_length-1].count('A') + \
                        new_third[j-(window_size//2):sequence_length-1].count('a') + \
                        new_third[0:rest].count('A') + \
                        new_third[0:rest].count('a')
                    c = new_third[j - (window_size // 2):sequence_length - 1].count('C') + \
                        new_third[j - (window_size // 2):sequence_length - 1].count('c') + \
                        new_third[0:rest].count('C') + \
                        new_third[0:rest].count('c')
                    g = new_third[j - (window_size // 2):sequence_length - 1].count('G') + \
                        new_third[j - (window_size // 2):sequence_length - 1].count('g') + \
                        new_third[0:rest].count('G') + \
                        new_third[0:rest].count('g')
                    t = new_third[j - (window_size // 2):sequence_length - 1].count('T') + \
                        new_third[j - (window_size // 2):sequence_length - 1].count('t') + \
                        new_third[0:rest].count('T') + \
                        new_third[0:rest].count('t')
                else:
                    a = new_third[j-(window_size//2):j + (window_size//2)].count('A') + \
                        new_third[j-(window_size//2):j + (window_size//2)].count('a')
                    c = new_third[j-(window_size//2):j + (window_size//2)].count('C') + \
                        new_third[j-(window_size//2):j + (window_size//2)].count('c')
                    g = new_third[j-(window_size//2):j + (window_size//2)].count('G') + \
                        new_third[j-(window_size//2):j + (window_size//2)].count('g')
                    t = new_third[j-(window_size//2):j + (window_size//2)].count('T') + \
                        new_third[j-(window_size//2):j + (window_size//2)].count('t')

                # Generates random gene position - plus or minus strand
                strand = np.random.choice([-1, 1])

                '''If not zero division, calculate and save the gc- and ta-skew'''
                if g + c != 0.0:
                    newsample_gc[j] = strand * (g-c)/(g+c)
                if t + a != 0.0:
                    newsample_ta[j] = strand * (t-a)/(t+a)

            # Find max (ter) and min (ori) positions in bootstrap sample
            ter_btstrap_gc = np.argmax(newsample_gc)
            ori_btstrap_gc = np.argmin(newsample_gc)

            ter_btstrap_ta = np.argmax(newsample_ta)
            ori_btstrap_ta = np.argmin(newsample_ta)

            '''Checks if the predicted positions are within the ranges'''
            if use_gc_ori and aroundori_gc[0] != 0 and aroundori_gc[-1] != len(newsample_gc) - 1:
                if aroundori_gc[0] <= ori_btstrap_gc <= aroundori_gc[-1]:
                    p_ori_gc = p_ori_gc + 1
            elif use_gc_ori and aroundori_gc[0] == 0 and aroundori_gc[-1] == len(newsample_gc) - 1:
                if ori_break_gc <= ori_btstrap_gc <= aroundori_gc[-1] or \
                        0 <= ori_btstrap_gc <= aroundori_gc[aroundori_gc.index(ori_break_gc) - 1]:
                    p_ori_gc = p_ori_gc + 1

            if use_gc_ter and aroundter_gc[0] != 0 and aroundter_gc[-1] != len(newsample_gc) - 1:
                if aroundter_gc[0] <= ter_btstrap_gc <= aroundter_gc[-1]:
                    p_ter_gc = p_ter_gc + 1
            elif use_gc_ter and aroundter_gc[0] == 0 and aroundter_gc[-1] == len(newsample_gc) - 1:
                if ter_break_gc <= ter_btstrap_gc <= aroundter_gc[-1] or \
                        0 <= ter_btstrap_gc <= aroundter_gc[aroundter_gc.index(ter_break_gc) - 1]:
                    p_ter_gc = p_ter_gc + 1

            if use_ta_ori and aroundori_ta[0] != 0 and aroundori_ta[-1] != len(newsample_ta) - 1:
                if aroundori_ta[0] <= ori_btstrap_ta <= aroundori_ta[-1]:
                    p_ori_ta = p_ori_ta + 1
            elif use_ta_ori and aroundori_ta[0] == 0 and aroundori_ta[-1] == len(newsample_ta) - 1:
                if ori_break_ta <= ori_btstrap_ta <= aroundori_ta[-1] or \
                        0 <= ori_btstrap_ta <= aroundori_ta[aroundori_ta.index(ori_break_ta) - 1]:
                    p_ori_ta = p_ori_ta + 1

            if use_ta_ter and aroundter_ta[0] != 0 and aroundter_ta[-1] != len(newsample_ta) - 1:
                if aroundter_ta[0] <= ter_btstrap_ta <= aroundter_ta[-1]:
                    p_ter_ta = p_ter_ta + 1
            elif use_ta_ter and aroundter_ta[0] == 0 and aroundter_ta[-1] == len(newsample_ta) - 1:
                if ter_break_ta <= ter_btstrap_ta <= aroundter_ta[-1] or \
                        0 <= ter_btstrap_ta <= aroundter_ta[aroundter_ta.index(ter_break_ta) - 1]:
                    p_ter_ta = p_ter_ta + 1

    '''If the p-values were not previously changed to 1.0, calculate 
    the final p-values'''
    if use_gc_ter:
        p_ter = p_ter_gc / N
    if use_gc_ori:
        p_ori = p_ori_gc / N
    if use_ta_ter:
        p_ter = p_ter_ta / N
    if use_ta_ori:
        p_ori = p_ori_ta / N

    return p_ori, p_ter


def main(record_dict, sequence, sequence_length, name, strain, output, graph=False):
    """
    Predicts the positions of ORI and TER based on GC skew and TA skew calculations.

    :param record_dict:     Dictionary storing all genes
    :param sequence:        FASTA sequence of the chromosome
    :param sequence_length: Length of the chromosome
    :param name:            Name of the species
    :param strain:          Strains of the species
    :param graph:           If graph should be produced
    :param output:          Specified output name and directory
    :return:                Predicted ORI and TER positions, list containing cumulative sums for
                            GC- and TA-skew, normalized lists of the values, perfect linear data
                            for both measurements given the predicted ORI and TER.
    """

    # Creating positions dictionary - stores all positions in all lists for all genes throughout the code
    positions = fill_positions_dict(record_dict)

    # Specifies which window size to start with
    window_size = int(0.00005*sequence_length)

    '''Produces the GC and TA skew lists'''
    gc_skew_list, ta_skew_list = prod_gc_skew(record_dict, positions, window_size, sequence_length, sequence)

    '''Find ter and ori'''
    ORI_gc, ORI_ta, TER_gc, TER_ta, gc_cum, ta_cum = find_ter_ori(gc_skew_list, ta_skew_list)

    '''Find actual positions -- assuming actual position for ORI is before the gene where the position is 
    indicated and that the actual position for TER is after the gene where the position is indicated.'''
    ORI_gc, ORI_gc_val, TER_gc, TER_gc_val, ORI_ta, ORI_ta_val, TER_ta, TER_ta_val = \
        find_actual_position(ORI_gc, ORI_ta, TER_gc, TER_ta, gc_cum, ta_cum, positions)

    '''Normalization --> All values between 0-1'''
    gc_normalized = minmax_scale(gc_cum, feature_range=(0, 1), axis=0, copy=True)
    ta_normalized = minmax_scale(ta_cum, feature_range=(0, 1), axis=0, copy=True)

    # Initate list for storing perfect, normalized data
    perfect_gc = []
    perfect_ta = []

    # X-axis for the perfect data
    x_perfect_gc = []
    x_perfect_ta = []

    '''Produce perfect plot from predicted ORI and TER'''
    if ORI_gc > TER_gc:
        perfect_gc, x_perfect_gc = perfect_plot_norm(TER_gc, ORI_gc, gc_normalized)
    elif TER_gc > ORI_gc:
        perfect_gc, x_perfect_gc = perfect_plot_norm(ORI_gc, TER_gc, gc_normalized)

    if ORI_ta > TER_ta:
        perfect_ta, x_perfect_ta = perfect_plot_norm(TER_ta, ORI_ta, ta_normalized)
    elif TER_ta > ORI_ta:
        perfect_ta, x_perfect_ta = perfect_plot_norm(ORI_ta, TER_ta, ta_normalized)

    return ORI_gc, TER_gc, gc_cum, gc_normalized, perfect_gc, x_perfect_gc, \
        ORI_ta, TER_ta, ta_cum, ta_normalized, perfect_ta, x_perfect_ta, positions


def around_pos(normalized, ORI, TER, av_diff):
    """
    Check which indices have a cum value differing less than the average difference

    :param normalized:  Normalized list of skew values
    :param ORI:         Predicted position of ORI
    :param TER:         Predicted position of TER
    :param av_diff:     Average difference between perfect and real data
    :return:            Lists of indices around predicted positions with value
                        differing less than the average difference,and the break points if found
    """

    # Initiating lists
    aroundter = list()
    aroundori = list()

    # Initiating break points
    ori_break = -1
    ter_break = -1

    # Length of normalized list
    len_normalized = len(normalized)

    # List of all indices
    indices = list(range(0, len_normalized))

    '''Saves all indices with values differing less than the average difference, 
    and saves the index as a break point when the indices jump more than one step'''
    for val in indices:
        if abs(normalized[val] - normalized[ORI-1]) < abs(av_diff):
            if len(aroundori) != 0:
                if val > aroundori[len(aroundori)-1] + 1:
                    ori_break = val
            aroundori.append(val)

        if abs(normalized[val] - normalized[TER-1]) < abs(av_diff):
            if len(aroundter) != 0:
                if val > aroundter[len(aroundter) - 1] + 1:
                    ter_break = val
            aroundter.append(val)

    '''Deletes all indices from list which are less than the break point if 
    the predicted position is above the break point, and deletes all indices
    from list which are greater than the break point, if the predicted position is 
    less than the break point - IF the first index is not the beginning of the sequence 
    and the last is not the end of the sequence.'''
    if ori_break != -1 and ORI >= ori_break and aroundori[0] != 0 \
            and aroundori[1] != len_normalized - 1:
        del aroundori[0:aroundori.index(ori_break)]
    elif ori_break != -1 and ORI < ori_break and aroundori[0] != 0 \
            and aroundori[1] != len_normalized - 1:
        del aroundori[aroundori.index(ori_break)::]

    if ter_break != -1 and TER >= ter_break and aroundter[0] != 0 \
            and aroundter[1] != len_normalized - 1:
        del aroundter[0:aroundter.index(ter_break)]
    elif ter_break != -1 and TER < ter_break and aroundter[0] != 0 \
            and aroundter[1] != len_normalized - 1:
        del aroundter[aroundter.index(ter_break)::]

    return list(aroundori), list(aroundter), ori_break, ter_break


def find_final_pred(ORI_gc, TER_gc, p_ori_gc, p_ter_gc, ORI_ta, TER_ta, p_ori_ta, p_ter_ta,
                    av_diff_gc, av_diff_ta, alpha):
    """
    Decides on final prediction between GC- and TA-skew.

    :param ORI_gc:      Predicted position of ORI by GC-skew
    :param TER_gc:      Predicted position of TER by GC-skew
    :param p_ori_gc:    p-value for ORI_gc
    :param p_ter_gc:    p-value for TER_gc
    :param ORI_ta:      Predicted position of ORI by TA-skew
    :param TER_ta:      Predicted position of TER by TA-skew
    :param p_ori_ta:    p-value for ORI_ta
    :param p_ter_ta:    p-value for TER_ta
    :param av_diff_gc:  Average difference between perfect and real gc-skew values
    :param av_diff_ta:  Average difference between perfect and real ta-skew values
    :param alpha:       Significance level
    :return:            Final predictions for ORI and TER and the final p-values
    """
    final_ORI = -1
    final_TER = -1
    final_p_ori = -1.0
    final_p_ter = -1.0

    '''If the p-values for the gc and ta positions are all below the significance level,
    checks which of the measurements have the least average difference. If the average difference 
    is the same, combines the two probabilities and the positions.'''
    if p_ori_gc <= alpha >= p_ter_gc and p_ori_ta <= alpha >= p_ter_ta:
        if av_diff_gc < av_diff_ta:
            final_ORI = ORI_gc
            final_TER = TER_gc
            final_p_ori = p_ori_gc
            final_p_ter = p_ter_gc
        elif av_diff_ta < av_diff_gc:
            final_ORI = ORI_ta
            final_TER = TER_ta
            final_p_ori = p_ori_ta
            final_p_ter = p_ter_ta
        elif av_diff_gc == av_diff_ta:
            final_ORI = (ORI_gc + ORI_ta)//2
            final_TER = (TER_gc + TER_ta)//2
            stat, final_p_ori = stats.combine_pvalues([p_ori_gc, p_ori_ta], method='fisher', weights=None)
            stat, final_p_ter = stats.combine_pvalues([p_ter_gc, p_ter_ta], method='fisher', weights=None)

    elif p_ori_gc <= alpha >= p_ter_gc and not (p_ori_ta <= alpha >= p_ter_ta):
        final_ORI = ORI_gc
        final_TER = TER_gc
        final_p_ori = p_ori_gc
        final_p_ter = p_ter_gc
    elif p_ori_ta <= alpha >= p_ter_ta and not (p_ori_gc <= alpha >= p_ter_gc):
        final_ORI = ORI_ta
        final_TER = TER_ta
        final_p_ori = p_ori_ta
        final_p_ter = p_ter_ta

    return final_ORI, final_TER, final_p_ori, final_p_ter



import warnings
import numpy as np
import argparse
import os
import errno
import tempfile
import subprocess
from Bio import SeqIO
import matplotlib.pyplot as plt
from collections import defaultdict
import seaborn as sns
from scipy import stats, signal, integrate
import time


warnings.filterwarnings("ignore")

# --- FUNCTIONS --- #


def sliding_sumdiff(sequence, win_size):
    """
    Iterate over an array of gene orientation values with a double-sided sliding window. For each iteration,
    calculate the difference between the sum of the right side and the sum of the left.

    Keyword arguments:
    sequence --
    win_size --
    :param sequence: the gene orientation array. Each element is either -1 or 1
    :param win_size: the size of each side of the double sliding window (int). Total window size is 2*win_size + 1
    :return: list with the result calculations with same length as input array
    """

    # Extend the data array to facilitate calculation of end positions
    sequence_extended = np.append(sequence, sequence[0:win_size])
    sequence_extended = np.append(sequence[-win_size:], sequence_extended)

    return [sum(sequence_extended[i + win_size: i + (2 * win_size)]) - sum(sequence_extended[i: i + win_size]) for i in
            range(len(sequence))]


def find_genes(in_file, length=False):
    """
    Run Prodigal to find genes in an input FASTA file. Save the result in a temporary file, parse it using Biopython
    SeqIO. Optionally read length of genome using a second temporary file via a FASTA output from Prodigal.

    :param in_file: the name of the input file
    :param length: whether or not to output genome length (boolean). Default is False
    :return: a dict with the genes calculated by Prodigal under 'genes'. Optionally also genome length under 'length'
    """

    print("Running Prodigal...\n")

    tmp1 = tempfile.NamedTemporaryFile(delete=False)
    prod_command = ["prodigal", '-i', in_file, '-f', 'gff', '-d', tmp1.name]
    if length:
        tmp2 = tempfile.NamedTemporaryFile(delete=False)
        prod_command = prod_command + ['-o', tmp2.name]  # add secondary output to Prodigal command to read length
    prodigal_out = None
    try:
        FNULL = open(os.devnull, 'w')
        subprocess.check_call(prod_command, stdout=FNULL, shell=False)
        prodigal_out = SeqIO.to_dict(SeqIO.parse(tmp1.name, 'fasta'))
    except subprocess.CalledProcessError:
        print("Error with Prodigal")
    tmp1.close()
    os.unlink(tmp1.name)
    if length:
        tmp2.close()
        with open(tmp2.name, 'r') as f:
            f.readline()
            gff_data = f.readline().rstrip()  # Read header line of secondary output

    ret = {'genes': prodigal_out}
    if length:
        ret['length'] = int(gff_data.split(';')[1].split('=')[1])  # Parse genome length from header

    os.unlink(tmp2.name)

    return ret


def parse_genes(prod_genes):
    """
    Parse the the Biopython SeqRecord representation of the Prodigal GFF output. Specifically, retrieve the
    start positions, end positions, strand orientation and sequences of each gene
    :param prod_genes: A SeqRecord object with genes found by Prodigal
    :return: A dict with the retrieved data
    """

    starts = []
    ends = []
    strands = []
    seqs = []
    for record in prod_genes:
        data = prod_genes[record].description.split(' # ')
        starts.append(int(data[1]))
        ends.append(int(data[2]))
        strands.append(int(data[3]))
        seqs.append(prod_genes[record].seq)
    samp_id = data[0] # simply use the last entry to retrieve the ID (same for whole data set)

    return {'start': starts, 'end': ends, 'strand': strands, 'sequence': seqs, "id": samp_id}


def predict_ori_ter(data, win_size=None, method="cumulative"):
    """
    Calculate strand orientation trend of a genome (cumulative or sumdiff), and predict ori and ter. If
    cumulative, assume ori as the global minimum and ter as the maximum - opposite for sumdiff
    :param data: an array of strand orientation data. Each element is -1 or 1
    :param win_size: The window for sumdiff calculations. Only required if method is 'sumdiff'
    :param method: Which method to use for trend calculation. Accepts 'cumulative' (default) or 'sumdiff'
    :return: strand_sum: the result of the trend calculation. ori_prediction: the index of predicted ori site.
    ter_prediction: the index of predicted ter site
    """

    if not method == "cumulative" and not method == "sumdiff":
        raise Exception("Invalid input method (cumulative or sumdiff are options, default is cumulative)")
    if method == "sumdiff" and win_size == None:
        raise Exception("param win_size must be set for sumdiff calculation")
    if method == "cumulative":
        strand_sum = np.cumsum(data)
        ori_prediction = np.argmin(strand_sum)
        ter_prediction = np.argmax(strand_sum)
    else:
        strand_sum = sliding_sumdiff(data, win_size)
        ori_prediction = np.argmax(strand_sum)
        ter_prediction = np.argmin(strand_sum)

    return strand_sum, ori_prediction, ter_prediction


def ori_gene_to_bp(ori_pred, ter_pred, gene_intervals):
    """
    Convert predicted ori and ter gene indices to a basepair position interval. Interval is from the end of the gene
    at prediction, to the start of the next gene in the genome.
    :param ori_pred: index of predicted ori site
    :param ter_pred: index of predicted ter site
    :param gene_intervals: a list of position intervals for all genes in the genome
    :return: ori_interval: the interval in which predicted ori lies. ter_interval: the interval in which predicted
    ter lies
    """

    ori_first = gene_intervals[ori_pred]
    ori_next = gene_intervals[0] if ori_pred == len(gene_intervals) - 1 else gene_intervals[ori_pred + 1]
    while ori_next[0] == ori_first[1]:
        ori_pred = ori_pred + 1
        ori_next = gene_intervals[0] if ori_pred == len(gene_intervals) - 1 else gene_intervals[ori_pred + 1]
    ori_interval = (ori_first[1], ori_next[0])

    ter_first = gene_intervals[ter_pred]
    ter_next = gene_intervals[0] if ter_pred == len(gene_intervals) - 1 else gene_intervals[ter_pred + 1]
    while ter_next[0] == ter_first[1]:
        ter_pred = ter_pred + 1
        ter_next = gene_intervals[0] if ter_pred == len(gene_intervals) - 1 else gene_intervals[ter_pred + 1]

    ter_interval = (ter_first[1], ter_next[0])

    return ori_interval, ter_interval


def find_best_prediction(data, gene_intervals, method="cumulative"):
    """
    Find the best prediction of ori and ter by iteratively decreasing window size and calculating position difference.
    When position changes less than 100 basepairs, accept as best solution.
    :param data: Array of strand orientation data. Each element is -1 or 1
    :param gene_intervals: The position intervals of all genes in the genome
    :param method: 'cumulative' (default) or 'sumdiff'
    :return: the trend calculation of best prediction, as well as the predicted gene indices and basepairs of ori and
    ter. Also returns the window size of the best prediction
    """

    min_win = len(data) // 6 if method == "sumdiff" else 2  # sumdiff method needs larger window than cumulative
    current_win = len(data) // 3 if method == "cumulative" else len(data) // 2  # see above comment
    current_sum, current_ori, current_ter = predict_ori_ter(data, win_size=current_win, method=method)
    current_ori_int, current_ter_int = ori_gene_to_bp(current_ori, current_ter, gene_intervals)

    while current_win >= min_win:
        new_win = current_win // 2
        new_sum, new_ori, new_ter = predict_ori_ter(data, new_win, method=method)
        new_ori_int, new_ter_int = ori_gene_to_bp(new_ori, new_ter, gene_intervals)
        if abs(new_ori_int[0] - current_ori_int[0]) < 100 and abs(new_ter_int[0] - current_ter_int[0]) < 100:
            break
        else:
            current_win = new_win
            current_sum = new_sum
            current_ori = new_ori
            current_ter = new_ter
            current_ori_int = new_ori_int
            current_ter_int = new_ter_int
    return current_sum, current_ori, current_ter, current_ori_int, current_ter_int, current_win


def position_distance(a, b, g_length):
    """
    Calculate the shortest distance between two positions on a circular genome.
    :param a: index of first position
    :param b: index of second position
    :param g_length: total genome length in basepairs
    :return: the calculated distance
    """

    diff = abs(a - b)
    return min(diff, g_length - diff)  # because circular, the shortest distance can be either a-b or length - (a-b)


def n_largest_smallest(data, n):
    """
    Find the indices of the n largest and smallest (respectively) values in an array.
    :param data: the data array
    :param n: the number of indices to find
    :return: two lists: one with the smallest indices and one with the largest
    """

    first_args = np.argpartition(data, n)[:n]
    last_args = np.argpartition(data, -n)[-n:]
    return [(arg, data[arg]) for arg in first_args], [(arg, data[arg]) for arg in last_args]


def perfect_plot(first, ori, ter, last):
    """
    Compute a "perfect" gene orientation plot given an ori and a ter position from a model data set (calculated trend).
     Consists of linear segments with a sign change around ori and ter.
    :param first: a tuple of the index and value of first position in model data
    :param ori: (index, value) of ori from model data
    :param ter: (index, value) of ter from model data
    :param last: (index, value) of the last position in model data
    :return: array with the "perfect" plot values
    """

    out_plot = np.asarray([])
    if ori[0] > ter[0]:
        if ori[0] != last[0]:
            out_plot = np.append(out_plot, np.linspace(start=ter[1], stop=ori[1], num=ori[0] - ter[0]))
            out_plot = np.append(out_plot, np.linspace(start=ori[1], stop=last[1], num=last[0] - ori[0] + 1))
        else:
            out_plot = np.append(out_plot, np.linspace(start=ter[1], stop=ori[1], num=ori[0] - ter[0] + 1))
        if ter[0] != first[0]:
            out_plot = np.append(np.linspace(start=first[1], stop=ter[1], num=ter[0] - first[0]), out_plot)
    elif ori[0] < ter[0]:
        if ori[0] != first[0]:
            out_plot = np.append(out_plot, np.linspace(start=ori[1], stop=ter[1], num=ter[0] - ori[0]))
            out_plot = np.append(np.linspace(start=first[1], stop=ori[1], num=ori[0] - first[0] + 1), out_plot)
        else:
            out_plot = np.append(out_plot, np.linspace(start=ori[1], stop=ter[1], num=ter[0] - ori[0]))
        if ter[0] != last[0]:
            out_plot = np.append(out_plot, np.linspace(start=ter[1], stop=last[1], num=last[0] - ter[0]))
    else:
        raise Exception("Impossible to predict graph (ori and ter are the same?!)")
    return out_plot


def real_perfect_diff(real_data, perfect_data):
    """
    Calculates the coefficient of variation (standard deviation / mean) of the pointwise absolute difference
    between real strand orientation trend data and its "perfect" representation as calculated with perfect_plot().
    :param real_data: array of the real data
    :param perfect_data: array of the perfect data
    :return: the coefficient of variation (float)
    """

    diff = [abs(perfect_data[i] - real_data[i]) for i in range(len(perfect_data))]
    return np.std(diff) / np.mean(diff)


def consecutive(data):
    """
    Find sections of data with consecutive indices in a list of tuples with (index, value) representing candidates for
    ori and ter
    :param data: list of tuples
    :return: a list of the middle tuples of each found section of consecutives
    """

    if len(data) == 1:
        return [data[0]]
    data.sort(key=lambda x: x[0])
    ret = []
    current_seq = [data[0]]
    for i in range(1, len(data)):
        if data[i][0] == data[i-1][0] + 1:
            current_seq.append(data[i])
        else:
            if len(current_seq) != 0:
                current_seq.append(data[i])
                ret.append(current_seq[(len(current_seq) - 1) // 2])
                current_seq = []
            else:
                current_seq.append(data[i])
    if len(current_seq) != 0:
        ret.append(current_seq[(len(current_seq) - 1) // 2])
    return ret


def outliers(data, cutoff_factor, side):
    """
    Find the upper or lower (determined by 'side') outliers in a list of (index, value) tuples representing
    ori or ter candidates.
    :param data: the input list of candidate tuples
    :param cutoff_factor: how many standard deviations to use as the outlier cutoff
    :param side: 'upper' or 'lower'
    :return: a list of the outlier tuples
    """

    values = [val[1] for val in data]
    data_mean, data_std = np.mean(values), np.std(values)
    cutoff = cutoff_factor * data_std
    if side == "lower":
        return [x for x in data if x[1] < data_mean - cutoff]
    elif side == "upper":
        return [x for x in data if x[1] > data_mean + cutoff]
    else:
        raise Exception("Invalid 'side' argument (lower or upper allowed)")


def find_unique_values(outliers):
    """
    Find unique values in a list of (index, value) tuples out ori and ter prediction outliers.
    :param outliers: An (index, value) tuple list of ori or ter prediction outliers
    :return: a list of (index, value) tuples of the middle index for each group of tuples sharing the same value
    """

    unique_dict = defaultdict(list)
    for val in outliers:
        if val[1] not in unique_dict.keys():
            unique_dict[val[1]] = []
        unique_dict[val[1]].append(val[0])

    return [(value[(len(value) - 1) // 2], key) for key, value in unique_dict.items()]


def select_ori_ter_candidates(data, n, initial_cutoff):
    """
    Find non-consecutive, unique, statistically outlying candidates of ori and ter. Start with a set number of the
    smallest and largest values of a trend calculation data sets, and reduce that candidate set through a series of
    methods.
    :param data: a data array of strand orientation trend calculation
    :param n: how many smallest and largest values to start with as candidates
    :param initial_cutoff: the initial standard deviation cutoff to determine statistical outliers
    :return: two lists of (index, value) tuples with ori and ter candidates
    """

    min_candidates, max_candidates = n_largest_smallest(data, n)
    min_outliers = []
    max_outliers = []

    # iteratively decrease stdev cutoff until at least one outlier (upper and lower) is found
    for i in range(initial_cutoff, 0, -1):
        min_outliers = min_outliers + outliers(min_candidates, i, "lower")
        max_outliers = max_outliers + outliers(max_candidates, i, "upper")
        if len(min_outliers) != 0 and len(max_outliers) != 0:
            break

    # if no outliers are found, move on with all initial candidates. Else, use the found outliers
    if len(min_outliers) == 0:
        min_ret = min_candidates
    else:
        min_ret = min_outliers
    if len(max_outliers) == 0:
        max_ret = max_candidates
    else:
        max_ret = max_outliers

    # only keep the candidates with unique values
    min_uniques = find_unique_values(min_ret)
    max_uniques = find_unique_values(max_ret)

    # only keep candidates with non-consecutive indices
    return consecutive(min_uniques), consecutive(max_uniques)


def combo_diffs(data, min_candidates, max_candidates):
    """
    For two lists of minimum and maximum candidates (respectively, ori and ter or vice versa), calculate the perfect
    plot for each possible combination and compute the coefficient of variation for the difference of that to the real
    data.
    :param data: The calculated trend data set
    :param min_candidates: the list of candidates for minimum position
    :param max_candidates: the list of candidates for maximum position
    :return: a list of tuples with (minimum index, maximum index, calculated coefficient of variation)
    """

    combos_diffs = []
    for min_val in min_candidates:
        for max_val in max_candidates:
            if min_val[0] != max_val[0]:  # if there are combinations with same index for both, ignore those
                perf_plot = perfect_plot((0, data[0]), min_val, max_val, (len(data) - 1, data[-1]))
                diff = real_perfect_diff(data, perf_plot)
                combos_diffs.append((min_val[0], max_val[0], diff))
    return combos_diffs


def go_find_ori(in_file, genes_dictionary, sequence_length):
    """
    Combine all previously defined functions to read a FASTA file and find the bestguess of ori and ter positions
    :param in_file:             File name of the input FASTA genome file
    :param genes_dictionary:    Dictionary containing all genes
    :return:                    The calculated trend data set, the perfect plot,
                                the ori interval and the ter interval of the best prediction,
                                list of gene start positions, list of gene end positions,
                                the method resulting in the best prediction, and the number of
                                plus and minus strands.
    """

    if not genes_dictionary:
        gene_out = find_genes(in_file, length=True)  # include calculation of genome length
        genes = gene_out['genes']
        gene_parse = parse_genes(genes)
    else:
        gene_parse = parse_genes(genes_dictionary)

    strands = gene_parse['strand']
    starts = gene_parse['start']
    ends = gene_parse['end']
    intervals = list(zip(starts, ends))

    strands = np.asarray(strands)

    strand_sum = np.cumsum(strands)
    strand_sumdiff = sliding_sumdiff(strands, len(strands) // 3)

    if sequence_length < 500000:
        no_max = 150
    else:
        no_max = 250

    cum_ori_cands, cum_ter_cands = \
        select_ori_ter_candidates(strand_sum, no_max, 5)        # start with 250 min and max candidates
    sd_ter_cands, sd_ori_cands = \
        select_ori_ter_candidates(strand_sumdiff, no_max, 5)    # and a stdev cutoff of 5

    cum_diffs = combo_diffs(strand_sum, cum_ori_cands, cum_ter_cands)
    cum_mean = np.mean([val[1] for val in cum_diffs])
    sd_diffs = combo_diffs(strand_sumdiff, sd_ori_cands, sd_ter_cands)
    sd_mean = np.mean([val[1] for val in sd_diffs])

    cum_best = [item for item in cum_diffs if item[2] == min([item[2] for item in cum_diffs])][0]
    sd_best = [item for item in sd_diffs if item[2] == min(item[2] for item in sd_diffs)][0]

    cum_best_dev = (cum_best[1] - cum_mean) / cum_mean
    sd_best_dev = (sd_best[1] - sd_mean) / sd_mean

    if cum_best_dev < sd_best_dev:
        final_prediction = cum_best
        best_ori = "cum-"
    elif cum_best_dev > sd_best_dev:
        final_prediction = sd_best
        best_ori = "sumdiff+"
    else:
        # if both predictions are equally good, decide with a coin flip
        if np.random.random(1) < 0.5:
            final_prediction = cum_best
            best_ori = "cum-"
        else:
            final_prediction = sd_best
            best_ori = "sumdiff+"

    if final_prediction == cum_best:
        data_plot = strand_sum
    else:
        data_plot = strand_sumdiff

    perf_plot = perfect_plot((0, data_plot[0]), (final_prediction[0], data_plot[final_prediction[0]]),
                             (final_prediction[1], data_plot[final_prediction[1]]), (len(strands) - 1, data_plot[-1]))

    ori_int, ter_int = ori_gene_to_bp(final_prediction[0], final_prediction[1], intervals)

    plus_content = round((len(np.where(np.equal(strands, 1))[0])/len(strands))*100)
    minus_content = 100 - plus_content

    return list(data_plot), list(perf_plot), list(ori_int), list(ter_int), starts, ends, best_ori,\
        plus_content, minus_content


def go_bootstrap(data_plot, perf_plot, starts, ends, ORI, TER, sequence_length, plus_content, minus_content,
                 log_file, best_ori = "cum-",
                 ori_rrna_range = list(), ter_rrna_range = list(), ori_go_rrna = list(), ter_go_rrna = list()):
    """
    Generates random lists with same number of genes as original sequence.
    The random lists are generated by randomly choosing -1 or 1 (weighted
    depending on the amount of -1 and 1 genes on original chromosome).
    :param data_plot:       List of data points generated by gene orientation function (cum or sumdiff)
    :param perf_plot:       Perfect (linear) plot of the data points given the ORI and TER positions
    :param starts:          List of gene start positions
    :param ends:            List of gene end positions
    :param ORI:             Predicted ORI interval
    :param TER:             Predicted TER interval
    :param sequence_length: Length of chromosome
    :param plus_content:    Number of plus strand genes on original sequence
    :param minus_content:   Number of minus strand genes on original sequence
    :param best_ori:        Method generating the best prediction
    :param ori_rrna_range:  Predicted ORI interval from rRNA
    :param ter_rrna_range:  Predicted TER interval from rRNA
    :return:                p-values for ORI and TER
    """

    # Lists containing indexes ranging from start to finish of the predicted rRNA intervals
    rrnas_ori = list()
    rrnas_ter = list()

    # Strand options - either plus or minus strand
    strand_opt = [-1, 1]

    # Weights - probability of either strand being randomly chosen
    weights = [minus_content / 100, plus_content / 100]

    # Predicted maximum distance from ORI and TER
    max_perc_genome = round(sequence_length * 0.70)
    min_perc_genome = int(sequence_length * 0.30)

    # Initiates p-values for ORI and TER
    p_ter_go = p_ori_go = float(0)

    '''If distance between ORI and TER interval starts or 
    ends is greater than the expected maximum distance, the prediction is likely to
    not be correct and bootstrap function is not performed.'''
    if not min_perc_genome <= abs(ORI[0] - TER[0]) <= max_perc_genome:
        p_ter_go = p_ori_go = 1.0
    else:
        # All indices are save to a list
        indices = list(range(0, len(data_plot)))

        # The indices of ORI and TER are saved as the end of the previous gene
        # and the beginning of the next gene
        if ORI[0] in ends:
            ind_ori = [ends.index(ORI[0]), starts.index(ORI[1])]
        elif ORI[1] in ends:
            ind_ori = [ends.index(ORI[1]), starts.index(ORI[0])]

        if TER[0] in ends:
            ind_ter = [ends.index(TER[0]), starts.index(TER[1])]
        elif TER[1] in ends:
            ind_ter = [ends.index(TER[1]), starts.index(TER[0])]

        # Initiates average difference between actual and perfect data
        av_diff_go = 0

        # Length of actual and perfect data lists
        len_perf_plot = len(perf_plot)
        len_data_plot = len(data_plot)

        '''Calculates average difference between actual and perfect data'''
        if len_data_plot == len_perf_plot:
            sum_diff_go = 0
            for i in range(len_perf_plot):
                sum_diff_go = sum_diff_go + abs(perf_plot[i] - data_plot[i])
            av_diff_go = abs(sum_diff_go / len_perf_plot)

        # Initiates list for storing indexes around ORI and TER with values
        # differing less than the average difference
        aroundter_go = list()
        aroundori_go = list()

        # Initiates index were there may be a jump in the plot. Eg when position
        # is predicted at end or beginning of plot, or otherwise noisy
        ori_break = -1
        ter_break = -1

        '''Saves all indices with value similar to the predicted origin and ter, 
        if there is a jump in indices, the jump positions are saved'''
        for val in indices:
            if abs(data_plot[val] - data_plot[ind_ori[0]]) < abs(av_diff_go) or \
                    abs(data_plot[val] - data_plot[ind_ori[1]]) < abs(av_diff_go):
                if len(aroundori_go) != 0:
                    if val > aroundori_go[len(aroundori_go)-1] + 1:
                        ori_break = val
                aroundori_go.append(val)
            if abs(data_plot[val] - data_plot[ind_ter[0]]) < abs(av_diff_go) or \
                    abs(data_plot[val] - data_plot[ind_ter[1]]) < abs(av_diff_go):
                if len(aroundter_go) != 0:
                    if val > aroundter_go[-1] + 1:
                        ter_break = val
                aroundter_go.append(val)

        '''If a break is found and the predicted origin is greater than the break position, 
        and the range does not span the end and beginning of the sequence, 
        the positions below the break point are deleted. 
        If a break is found and the predicted origin is less than the break position, 
        and the range does not span the end and beginning of the sequence, 
        the positions above the break point are deleted. '''
        if ori_break != -1 and ind_ori[0] >= ori_break and aroundori_go[0] != 0 and aroundori_go[-1] != len_data_plot - 1:
            del aroundori_go[0:aroundori_go.index(ori_break)]
        elif ori_break != -1 and ind_ori[1] <= ori_break and aroundori_go[0] != 0 and aroundori_go[-1] != len_data_plot - 1:
            del aroundori_go[aroundori_go.index(ori_break)::]

        if ter_break != -1 and ind_ter[0] >= ter_break and aroundter_go[0] != 0 and aroundter_go[-1] != len_data_plot - 1:
            del aroundter_go[0:aroundter_go.index(ter_break)]
        elif ter_break != -1 and ind_ter[1] <= ter_break and aroundter_go[0] != 0 and aroundter_go[-1] != len_data_plot - 1:
            del aroundter_go[aroundter_go.index(ter_break)::]

        # Initiates positions of bootstrap predictions
        ter_btstrap_go = 0
        ori_btstrap_go = 0

        #
        ori_start = ori_end = ter_start = ter_end = 0
        p_diff = 0.0

        # Sets number of iterations to 100 000
        N = 1000
        if best_ori == "sumdiff+":
            N = 1000

        with open(log_file, "a") as logs:
            logs.write("%s      %s" % ("    ", ".."))

        btstrap_go = list()
        for i in range(N):

            if i % 100 == 0:
                with open(log_file, "a") as logs:
                    logs.write("{}..".format(i))

            # Generates random sequence of genes
            new_sample_go = np.random.choice(strand_opt, len_data_plot, p=weights, replace=True)

            '''Checking which method generated the best prediction, 
            and decides whether ori and ter is in the maximum or minimum.'''
            if best_ori == "cum-":
                new_sample_go = np.cumsum(new_sample_go)
                ter_btstrap_go = np.argmax(new_sample_go)
                ori_btstrap_go = np.argmin(new_sample_go)
            elif best_ori == "sumdiff+":
                new_sample_go = sliding_sumdiff(new_sample_go, len(new_sample_go) // 3)
                ter_btstrap_go = np.argmin(new_sample_go)
                ori_btstrap_go = np.argmax(new_sample_go)

            if ter_btstrap_go == 0:
                ter_start = ter_start + 1
            elif ter_btstrap_go == len(new_sample_go) - 1:
                ter_end = ter_end + 1
            if ori_btstrap_go == 0:
                ori_start = ori_start + 1
            elif ori_btstrap_go == len(new_sample_go) - 1:
                ori_end = ori_end + 1

            btstrap_go.append([ori_btstrap_go, ter_btstrap_go])

            # if abs(ends[ori_btstrap_go]-starts[ter_btstrap_go]) <= sfperc_genome:
            #     p_diff = p_diff + 1

        '''Checks if the bootstrap predicted ORIs and TERs are within the ranges defined above.
        If yes --> p is increased'''
        for pos in btstrap_go:
            if ori_break != -1 and aroundori_go[0] == 0 and aroundori_go[-1] == len_data_plot - 1:
                if aroundori_go[0] <= pos[0] <= aroundori_go[aroundori_go.index(ori_break) - 1] or \
                        ori_break <= pos[0] <= aroundori_go[-1]:
                    p_ori_go = p_ori_go + 1
            elif ori_break == -1 and aroundori_go[0] != 0 and aroundori_go[-1] != len_data_plot - 1:
                if aroundori_go[0] <= pos[0] <= aroundori_go[-1]:
                    p_ori_go = p_ori_go + 1
            elif ori_break == -1 and aroundori_go[0] == 0 and aroundori_go[-1] == len_data_plot - 1:
                p_ori_go = p_ori_go + 1
            else:
                if aroundori_go[0] <= pos[0] <= aroundori_go[-1]:
                    p_ori_go = p_ori_go + 1

            if ter_break != -1 and aroundter_go[0] == 0 and aroundter_go[-1] == len_data_plot - 1:
                if aroundter_go[0] <= pos[1] <= aroundter_go[aroundter_go.index(ter_break) - 1] or \
                        ter_break <= pos[1] <= aroundter_go[-1]:
                    p_ter_go = p_ter_go + 1
            elif ter_break == -1 and aroundter_go[0] != 0 and aroundter_go[-1] != len_data_plot - 1:
                if aroundter_go[0] <= pos[1] <= aroundter_go[-1]:
                    p_ter_go = p_ter_go + 1
            elif ter_break == -1 and aroundter_go[0] == 0 and aroundter_go[-1] == len_data_plot - 1:
                p_ter_go = p_ter_go + 1
            else:
                if aroundter_go[0] <= pos[1] <= aroundter_go[-1]:
                    p_ter_go = p_ter_go + 1

        p_ter_go = p_ter_go / N
        p_ori_go = p_ori_go / N

        '''If the predicted ORI and TER are not within the rRNA range, 
        this is not a likely prediction --> p = 1.0'''
        if len(ori_rrna_range) != 0:
            if len(ori_go_rrna) == 0:
                p_ori_go = 1.0
        if len(ter_rrna_range) != 0:
            if len(ter_go_rrna) == 0:
                p_ter_go = 1.0

    return p_ori_go, p_ter_go



""""------------------------------------------------PACKAGES---------------------------------------------------------"""
from __future__ import print_function
from Bio import SeqIO
import subprocess
import errno
import tempfile
from teritori import Motifs as mot, GOscript as go, GCscript as gc, final_pred as fp
import argparse
from pylab import *
import csv
import numpy as np
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import os
from sklearn.preprocessing import minmax_scale


"""---------------------------------------------------------------------------------------------------------------------
---------------------------------------------------FUNCTIONS---------------------------------------------------------"""


def run_prodigal(input_file, temp_file):
    """
    Runs prodigal

    :param input_file:  Sequence FASTA file
    :param temp_file:   Name of temporary file
    :return:            Saves output to temporary file
    """
    try:
        prodigal = subprocess.Popen(['prodigal', '-i', input_file, '-d', temp_file, '-f', 'gff'],
                                    stdout=subprocess.PIPE, shell=False)
        prodigal.stdout.read().decode('utf-8')
    except subprocess.CalledProcessError:
        print("Error with Prodigal")


def main():
    """----------------------------------------------PARSER----------------------------------------------------------"""
    parser = argparse.ArgumentParser()

    parser.add_argument("-i", "--input", help="Genome file, FASTA format.")
    parser.add_argument("--genes", help="File containing genes from Prodigal")
    parser.add_argument("-o", "--output", help="Prefix for all files created in script", type=str)
    parser.add_argument("-l", "--linear", help="Specifies that the genome is linear",
                        action="store_true", default=False)
    parser.add_argument("-a", "--all", help="Specifies that all features should be used",
                        action="store_true", default=True)
    parser.add_argument("--gc", help="Use GC-script", action="store_true", default=False)
    parser.add_argument("--go", help="Use Gene orientation script", action="store_true", default=False)
    parser.add_argument("--cm", help="Use script for conserved motifs", action="store_true", default=False)
    parser.add_argument("--graph", help="Should graph be displayed or not?", action="store_true", default=False)
    args = parser.parse_args()

    if args.cm or args.go or args.gc:
        args.all = False

    """----------------------------------------FILE HANDLING---------------------------------------------------------"""
    # Current directory
    cwd = os.getcwd()

    # Path of the directory to be created
    outer_path = cwd + "/terITori"
    results_path = outer_path + "/Results"

    '''Creating directory if it does not already exist'''
    if not os.path.isdir(outer_path):
        try:
            os.mkdir(outer_path)
        except OSError:
            sys.exit("Creation of directory %s failed" % outer_path)
    if not os.path.isdir(results_path):
        try:
            os.mkdir(results_path)
        except OSError:
            sys.exit("Creation of directory %s failed" % results_path)

    # Defines the input file
    in_file = args.input
    gene_file = ""
    if not os.path.exists(os.path.abspath(in_file)):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), in_file)

    # Opens file. The file is properly closed after reading it
    # "rU" --> reading using universal readline mode - independent of OS
    with open(in_file, "rU") as file:
        sequence = SeqIO.read(file, "fasta")     # Reads the sequence, saves sequence in variable.

    """--------------------------------------UNIVERSAL VARIABLES-----------------------------------------------------"""
    # Saves species info
    split_desc = sequence.description.split(",")[0]                         # Full title
    sequence_length = len(sequence.seq)                                     # Full sequence length
    name_species = split_desc.split(None)[1] + " " + split_desc.split(None)[2]      # Name
    strain_split = split_desc.split(None)[3::]                              # Strain information
    strain = ""
    for i in range(0, len(strain_split)):                           # Strain information as string
        strain = strain + strain_split[i] + " "

    # Significance level
    alpha = 0.05

    # Linear or not
    linear = args.linear

    """---------------------------------CREATE RESULTS FOLDER AND FILE-----------------------------------------------"""
    # Create path for results'''
    species = name_species.replace(" ", "_") if not args.output else args.output # uses user-specified prefix as name if it is specified
    inner_path = results_path + "/" + species
    if not os.path.isdir(inner_path):
        try:
            os.mkdir(inner_path)
        except OSError:
            sys.exit("Creation of inner directory %s failed" % inner_path)

    # Create file for handling results'''
    csv_file = inner_path + "/" + species + "_results_teritori.csv"

    with open(csv_file, mode='w') as results:
        teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
        teritori_writer.writerow([species, "Sequence length:", sequence_length, '', '', ''])
        teritori_writer.writerow(['Prediction measurement', 'ORI', 'p-value \n(ori)', 'TER', 'p-value \n(ter)',
                                  'Comment'])

    log_file = inner_path + "/" + species + "_log.txt"

    with open(log_file, "w") as logs:
        logs.write("{}       Log file -- terITori    Created: {}\n\n".format(
            species, time.asctime(time.localtime(time.time()))))

    """-----------------------------------------------PRODIGAL-------------------------------------------------------"""
    # Dictionary for storing gene information
    genes = list()
    record_dict = dict()

    '''Runs Prodigal if gene orientation, gc-skew or all have been chosen in input, 
    and if a gene file has not been entered'''
    if args.go or args.gc or args.all:
        # Creating temporary file
        temp = tempfile.NamedTemporaryFile(delete=True)

        temp.name = "genes_" + species

        '''If a gene file is entered when running the program, the genes are collected from file.
        Else, Prodigal is run'''
        if args.genes:
            with open(log_file, "a") as logs:
                logs.write("%s      %s      %s" % (time.asctime(time.localtime(time.time())),
                                                   "Genes entered as input -- not running Prodigal.\n", ""))

            gene_file = os.path.abspath(args.genes)
            genes = list(SeqIO.parse(gene_file, "fasta"))

        elif not args.genes:
            ''' Prodigal - returns genes'''
            with open(log_file, "a") as logs:
                logs.write("%s      %s     " % (time.asctime(time.localtime(time.time())),
                                                "No genes previously found - Running Prodigal..."))
                run_prodigal(in_file, temp.name)
                gene_file = os.path.abspath(temp.name)

                logs.write("%s" % "Prodigal successful!\n")

        # Puts all genes from Prodigal or previous gene file into a dictionary'''
        record_dict = dict(SeqIO.index(gene_file, "fasta"))

        '''Saves the dictionary as a file if genes were not specified in input'''
        genes_directory = inner_path + "/" + temp.name
        if not args.genes:
            with open(genes_directory, 'w') as f:
                for key, value in record_dict.items():
                    f.write('%s%s\n' % (">" + value.description, "\n" + value.seq))

        # Closes temporary file'''
        temp.close()

    """-----------------------------------------------------------------------------------------------------------------
    -------------------------------------------CONSERVED MOTIFS---------------------------------------------------------
    -----------------------------------------------------------------------------------------------------------------"""
    # Coordinates from rRNA prediction
    ORI_rrnas = [0, 0]
    TER_rrnas = [0, 0]

    # Coordinates from dnaA and dif motif predictions
    ori_dnaa_boxes = [0, 0]
    ter_dif_motifs = [0, 0]

    # Index where the most number of dnaA boxes are present
    # Used to validate ori_dnaa_boxes interval
    dnaa_max_pos = 0
    dif_max_pos = 0

    # List for storing rRNA results
    rrnas = list()

    # Lists for storing all indices in intervals
    ori_rrna_range = list()
    ter_rrna_range = list()
    ori_dnaa_range = list()
    ter_dif_range = list()

    # Lists for storing sorted dnaA boxes and dif motifs
    dnaa_sorted = list()
    dif_sorted = list()

    '''If -cm is specified: Runs Motifs.py script'''
    if args.cm or args.all:
        with open(log_file, "a") as logs:
            logs.write("%s      %s" % (time.asctime(time.localtime(time.time())), "Starting Conserved Motifs...\n"))

        with open(log_file, "a") as logs:
            logs.write("%s      %s                           " % (time.asctime(time.localtime(time.time())),
                                                                  "...Searching for rRNAs..."))

        ''' ---FINDING rRNAs--- '''
        rrnas = mot.finding_rrnas(in_file)
        '''If rRNAs are found, find ORI and TER and save to file'''
        if len(rrnas) != 0:
            ORI_rrnas = mot.finding_ori(rrnas, sequence_length)
            TER_rrnas = mot.finding_ter(rrnas, sequence_length)
            if ORI_rrnas == [0, 0] == TER_rrnas:
                with open(csv_file, mode='a') as results:
                    teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                    teritori_writer.writerow(["rRNAs", "---", "", "---", "", "Predictions not possible"])
            elif ORI_rrnas == [0, 0] != TER_rrnas:
                with open(csv_file, mode='a') as results:
                    teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                    teritori_writer.writerow(["rRNAs", "---", "", TER_rrnas, "", "ORI prediction not possible"])
            elif ORI_rrnas != [0, 0] == TER_rrnas:
                with open(csv_file, mode='a') as results:
                    teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                    teritori_writer.writerow(["rRNAs", ORI_rrnas, "", "---", "", "TER prediction not possible"])
            else:
                with open(csv_file, mode='a') as results:
                    teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                    teritori_writer.writerow(["rRNAs", ORI_rrnas, "", TER_rrnas, "", ""])

            if ORI_rrnas != [0, 0]:
                if ORI_rrnas[1] < ORI_rrnas[0]:
                    ori_rrna_range = list(range(ORI_rrnas[1], sequence_length-1)) + \
                                     list(range(0, ORI_rrnas[0]))
                elif ORI_rrnas[0] <= ORI_rrnas[1]:
                    ori_rrna_range = list(range(ORI_rrnas[0], ORI_rrnas[1]))

            if TER_rrnas != [0, 0]:
                if TER_rrnas[1] < TER_rrnas[0]:
                    ter_rrna_range = list(range(TER_rrnas[1], sequence_length -1)) + \
                                     list(range(0, TER_rrnas[0]))
                elif TER_rrnas[0] <= TER_rrnas[1]:
                    ter_rrna_range = list(range(TER_rrnas[0], TER_rrnas[1]))
        else:
            with open(csv_file, mode='a') as results:
                teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                teritori_writer.writerow(["rRNAs", "---", "", "---", "", "No rRNAs found, prediction not possible"])

        # rRNAs finished
        with open(log_file, "a") as logs:
            logs.write("%s" % "Done!\n")

        ''' ---FINDING dnaA BOXES--- '''
        with open(log_file, "a") as logs:
            logs.write("%s      %s                      " % (time.asctime(time.localtime(time.time())),
                                                             "...Searching for dnaA boxes..."))

        '''Runs nhmmer to find dnaA boxes'''
        mot.hmm_runner(in_file, os.path.dirname(os.path.abspath( __file__ )) + '/conserved_motifs/dnaaboxes.hmm', os.path.dirname(os.path.abspath( __file__ )) + '/prefix_dnaa_boxes.txt')

        '''Parses the output from nhmmer, containing dnaA boxes'''
        dnaa_boxes = mot.hmm_parser(os.path.dirname(os.path.abspath( __file__ )) + '/prefix_dnaa_boxes.txt')

        '''If dnaA boxes are found, determines ORI'''
        if len(dnaa_boxes) != 0:
            '''Finds the positions of Ori based on dnaA boxes'''
            ori_dnaa_boxes, dnaa_sorted = mot.finding_dnaa_boxes(dnaa_boxes, ORI_rrnas[0], ORI_rrnas[1])

            '''If the interval spans the end and beginning of the sequence,
            define all intervals including from beginning of interval to end of sequence
            and from beginning of sequence to end of interval. 
            Otherwise, from beginning of interval to end of interval'''
            if ori_dnaa_boxes != [0, 0]:
                if ori_dnaa_boxes[1] < ori_dnaa_boxes[0]:
                    ori_dnaa_range = list(range(ori_dnaa_boxes[1], sequence_length-1)) + \
                                     list(range(0, ori_dnaa_boxes[0]))
                elif ori_dnaa_boxes[0] <= ori_dnaa_boxes[1]:
                    ori_dnaa_range = list(range(ori_dnaa_boxes[0], ori_dnaa_boxes[1]))

                if len(ori_dnaa_range) > sequence_length//3:
                    ori_dnaa_range = list()

        '''Saving results from dnaA prediction'''
        if ori_dnaa_boxes != [0, 0]:
            with open(csv_file, mode='a') as results:
                teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                teritori_writer.writerow(["dnaA", ori_dnaa_boxes, "", "---", "", "Prediction of ORI"])
        else:
            with open(csv_file, mode='a') as results:
                teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                teritori_writer.writerow(["dnaA", "---", "", "---", "",
                                          "No significant dnaA boxes found, prediction not possible"])
        # dnaA interval finished
        with open(log_file, "a") as logs:
            logs.write("%s" % "Done!\n")

        ''' ---FINDING dnaA MAXIMUM POSITION--- '''
        with open(log_file, "a") as logs:
            logs.writelines("%s      %s                  " % (time.asctime(time.localtime(time.time())),
                                                              "...Calculating dnaA box maximum..."))

        '''If dnaA boxes are found, 1000 bp window with the most number of dnaa boxes'''
        if len(dnaa_boxes) != 0 and ORI_rrnas != [0, 0]:
            dnaa_max_pos = mot.dnaa_max(dnaa_boxes, ORI_rrnas, sequence_length, name_species)

            '''If the index is within the dnaA ORI interval, trust the interval.'''
            if len(ori_dnaa_range) != 0:
                '''Checks if the index is within the interval and saves result.'''
                if dnaa_max_pos in ori_dnaa_range:
                    with open(csv_file, mode='a') as results:
                        teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                        teritori_writer.writerow(["dnaA max", dnaa_max_pos, "", "---", "",
                                                  "Prediction in dnaA prediction range"])
                else:
                    with open(csv_file, mode='a') as results:
                        teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                        teritori_writer.writerow(['dnaA max', dnaa_max_pos, "", "---", "",
                                                  "OBS! Prediction outside dnaA prediction range"])
        # dnaA box maximum finished
        with open(log_file, "a") as logs:
            logs.write("%s" % "Done!\n")

        ''' ---FINDING DIF MOTIFS--- '''
        with open(log_file, "a") as logs:
            logs.write("%s      %s" % (time.asctime(time.localtime(time.time())),
                                       "...Finding dif motifs...\n".format()))

        '''Runs nhmmer to find dif motifs'''
        mot.hmm_runner(in_file, os.path.dirname(os.path.abspath(__file__)) + '/conserved_motifs/difmotifs.hmm', os.path.dirname(os.path.abspath(__file__)) + '/prefix_dif_motifs.txt')

        '''Parses the output from nhmmer, containing dif motifs'''
        dif_motifs = mot.hmm_parser(os.path.dirname(os.path.abspath(__file__)) + '/prefix_dif_motifs.txt')

        '''If dif motifs are found, determine TER interval'''
        if len(dif_motifs) != 0:
            '''Finds the positions of Ter based on dif motifs'''
            ter_dif_motifs, dif_sorted = mot.finding_dif_motifs(dif_motifs, TER_rrnas[0], TER_rrnas[1])

            if ter_dif_motifs != [0, 0]:
                if ter_dif_motifs[1] < ter_dif_motifs[0]:
                    ter_dif_range = list(range(ter_dif_motifs[1], sequence_length - 1)) + \
                                    list(range(0, ter_dif_motifs[0]))
                elif ter_dif_motifs[0] <= ter_dif_motifs[1]:
                    ter_dif_range = list(range(ter_dif_motifs[0], ter_dif_motifs[1]))

                if len(ter_dif_range) > sequence_length//3:
                    ter_dif_range = list()

        '''Saving results from dif prediction'''
        if ter_dif_motifs != [0, 0]:
            with open(csv_file, mode='a') as results:
                teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                teritori_writer.writerow(["dif", "---", "", ter_dif_motifs, "", "Prediction of TER"])
        else:
            with open(csv_file, mode='a') as results:
                teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                teritori_writer.writerow(["dif", "---", "", "---", "",
                                          "No significant dif motifs found, prediction not possible"])

            if ter_dif_motifs != [0, 0]:
                if ter_dif_motifs[1] < ter_dif_motifs[0]:
                    dif_range = list(range(ter_dif_motifs[1], sequence_length)) + \
                                list(range(0, ter_dif_motifs[0]))
                elif ter_dif_motifs[0] <= ter_dif_motifs[1]:
                    dif_range = list(range(ter_dif_motifs[0], ter_dif_motifs[1]))

        # dif motifs finished
        with open(log_file, "a") as logs:
            logs.write("%s" % "Done!\n")

        ''' ---FINDING DIF MOTIF MAXIMUM--- '''
        with open(log_file, "a") as logs:
            logs.writelines("%s     %s                  " % (time.asctime(time.localtime(time.time())),
                                                             "...Calculating dif motif maximum..."))

        '''If dnaA boxes are found, 1000 bp window with the most number of dnaa boxes'''
        if len(dif_motifs) != 0 and TER_rrnas != [0, 0]:
            dif_max_pos = mot.dnaa_max(dif_motifs, TER_rrnas, sequence_length, name_species)

            '''If the index is within the dnaA ORI interval, trust the interval.'''
            if len(ter_dif_range) != 0:
                '''Checks if the index is within the interval and saves result.'''
                if dif_max_pos in ter_dif_range:
                    with open(csv_file, mode='a') as results:
                        teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                        teritori_writer.writerow(["dif max", "", "", dif_max_pos, "",
                                                  "Prediction in dif prediction range"])
                else:
                    with open(csv_file, mode='a') as results:
                        teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                        teritori_writer.writerow(['dif max', "", "", dif_max_pos, "",
                                                  "OBS! Prediction outside dif prediction range"])
        # dif motif maximum finished
        with open(log_file, "a") as logs:
            logs.write("%s" % "Done!\n")

        # Motifs completely finished
        with open(log_file, "a") as logs:
            logs.write("%s      %s                                                   %s"
                       % (time.asctime(time.localtime(time.time())), " ", "Motifs prediction finished! \n"))

    """-----------------------------------------------------------------------------------------------------------------
    -----------------------------------------------GENE ORIENTATION-----------------------------------------------------
    -----------------------------------------------------------------------------------------------------------------"""
    # Coordinates for ORI and TER
    ORI_go = [0, 0]
    TER_go = [0, 0]

    # Gene orientation p-values
    p_ori_go = p_ter_go = 1.0

    # Lists for saving results and overlaps
    ori_go_range = list()
    ori_go_rrna = list()
    ori_go_dnaa = list()
    ter_go_range = list()
    ter_go_rrna = list()
    ter_go_dif = list()

    '''If -go is specified: Runs GOscript.py script'''
    if args.go or args.all:
        with open(log_file, "a") as logs:
            logs.write("%s      %s" % (time.asctime(time.localtime(time.time())), "Starting Gene Orientation...\n"))

        '''Finds the ORI and TER using gene orientation calculations'''
        data_plot, perf_plot, ORI_go, TER_go, starts, ends, best_ori, plus_content, minus_content \
            = go.go_find_ori(in_file, record_dict, sequence_length)

        with open(log_file, "a") as logs:
            logs.write("%s      %s" % (time.asctime(time.localtime(time.time())),
                                       "...Gene orientation data calculated.\n"))

        # Length of produced list
        len_data_plot = len(data_plot)

        '''The list of values is normalized using the value 
        which has the greatest absolute value'''
        if abs(max(data_plot)) > abs(min(data_plot)):
            div = abs(max(data_plot))
        else:
            div = abs(min(data_plot))
        normalized_go = minmax_scale(data_plot, feature_range=(0, 1), axis=0, copy=True)
        normalized_go_perf = minmax_scale(perf_plot, feature_range=(0, 1), axis=0, copy=True)

        '''If the start of the interval is greater than the end of the interval, 
        the intervals are swapped.'''
        if ORI_go[0] > ORI_go[1]:
            ORI_go = list(reversed(ORI_go))
        if TER_go[0] > TER_go[1]:
            TER_go = list(reversed(TER_go))

        ori_go_range = list(range(ORI_go[0], ORI_go[1]))
        ter_go_range = list(range(TER_go[0], TER_go[1]))

        '''If the positions have been previously predicted using the motifs, 
        the gene orientation bootstrap function is run using the previously predicted intervals'''
        if args.cm or args.all:
            '''If dnaA and dif intervals were found, use these for predicting the probabilities of the gene orientation
            interval'''
            if len(ori_dnaa_range) != 0 and len(ter_dif_range) != 0:
                # Saves all positions with intervals
                ori_go_dnaa = list(set(ori_go_range) & set(ori_dnaa_range))
                ter_go_dif = list(set(ter_go_range) & set(ter_dif_range))

                '''Checks if gene orientation intervals overlap with dnaA and dif intervals
                and in that case uses these intervals to predict prediction probability. 
                Otherwise the gene orientation prediction is not probable'''
                if len(ori_go_dnaa) != 0 or len(ter_go_dif) != 0:
                    with open(log_file, "a") as logs:
                        logs.write("%s      %s      " % (time.asctime(time.localtime(time.time())),
                                                         "...Starting Bootstrap function...\n"))

                    p_ori_go, p_ter_go = go.go_bootstrap(normalized_go, normalized_go_perf, starts, ends,
                                                         ORI_go, TER_go, sequence_length, plus_content, minus_content,
                                                         log_file, best_ori,
                                                         ori_dnaa_range, ter_dif_range, ori_go_dnaa, ter_go_dif)
                    with open(log_file, "a") as logs:
                        logs.write("%s" % "   Done! \n")

                else:
                    with open(log_file, "a") as logs:
                        logs.write("%s      %s" % (time.asctime(time.localtime(time.time())),
                                                   "Gene orientation predictions not within probable range. "
                                                   "Bootstrap not performed.\n"))
                    p_ori_go = 1.0
                    p_ter_go = 1.0

            elif len(ori_rrna_range) != 0 and len(ter_rrna_range) != 0:
                # Saves the overlap between gene orientation interval and rRNA intervals
                ori_go_rrna = list(set(ori_go_range) & set(ori_rrna_range))
                ter_go_rrna = list(set(ter_go_range) & set(ter_rrna_range))

                '''Checking if gene orientation interval is within the rrna interval, 
                and uses this to evaluate the uncertainty of the predictions. Otherwise
                the gene orientation prediction is not probably.'''
                if len(ori_go_rrna) != 0 or len(ter_go_rrna) != 0:
                    with open(log_file, "a") as logs:
                        logs.write("%s      %s                   " % (time.asctime(time.localtime(time.time())),
                                                                      "...Starting Bootstrap function...\n"))

                    p_ori_go, p_ter_go = go.go_bootstrap(normalized_go, normalized_go_perf, starts, ends,
                                                         ORI_go, TER_go, sequence_length, plus_content, minus_content,
                                                         log_file, best_ori,
                                                         ori_rrna_range, ter_rrna_range, ori_go_rrna, ter_go_rrna)
                    with open(log_file, "a") as logs:
                        logs.write("%s" % "   Done! \n")
                else:
                    with open(log_file, "a") as logs:
                        logs.write("%s      %s" % (time.asctime(time.localtime(time.time())),
                                                   "Gene orientation predictions not within probable range. "
                                                   "Bootstrap not performed.\n"))
                    p_ori_go = 1.0
                    p_ter_go = 1.0
            else:
                with open(log_file, "a") as logs:
                    logs.write("%s      %s                   " % (time.asctime(time.localtime(time.time())),
                                                                  "...Starting Bootstrap function...\n"))

                p_ori_go, p_ter_go = go.go_bootstrap(normalized_go, normalized_go_perf, starts, ends,
                                                     ORI_go, TER_go, sequence_length, plus_content, minus_content,
                                                     log_file, best_ori)
                with open(log_file, "a") as logs:
                    logs.write("%s" % "   Done! \n")
        else:
            with open(log_file, "a") as logs:
                logs.write("%s      %s                   " % (time.asctime(time.localtime(time.time())),
                                                              "...Starting Bootstrap function...\n"))

            p_ori_go, p_ter_go = go.go_bootstrap(normalized_go, normalized_go_perf, starts, ends,
                                                 ORI_go, TER_go, sequence_length, plus_content, minus_content,
                                                 log_file, best_ori)

            with open(log_file, "a") as logs:
                logs.write("%s" % "   Done! \n")

        comment = ""
        '''Saves Gene orientation predictions to file'''
        if ORI_go != [0, 0] and TER_go != [0, 0]:
            if p_ori_go > alpha:
                comment = "NB! ORI p-value above significance level, \nORI prediction should not be trusted! "
            if p_ter_go > alpha and comment == "":
                comment = "NB! TER p-value above significance level, \nTER prediction should not be trusted! "
            elif p_ter_go > alpha and comment != "":
                comment = "NB! Both p-values above signficance level, \nGene orientation should not be trusted! "

            with open(csv_file, mode='a') as results:
                teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                teritori_writer.writerow(["Gene orientation", ORI_go, p_ori_go, TER_go, p_ter_go, comment])

        elif ORI_go != [0, 0] and TER_go == [0, 0]:
            if p_ori_go > alpha:
                comment = "NB! p-value above significance level, \nORI prediction should not be trusted! "

            with open(csv_file, mode='a') as results:
                teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                teritori_writer.writerow(["Gene orientation", ORI_go, p_ori_go, "---", "",
                                          comment + "TER prediction not possible."])
        elif ORI_go == [0, 0] and TER_go != [0, 0]:
            if p_ter_go > alpha:
                comment = "NB! p-value above significance level, \nTER prediction should not be trusted! "

            with open(csv_file, mode='a') as results:
                teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                teritori_writer.writerow(["Gene orientation", "---", "", TER_go, p_ter_go,
                                          comment + "ORI prediction not possible."])
        else:
            with open(csv_file, mode='a') as results:
                teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
                teritori_writer.writerow(["Gene orientation", "---", "", "---", "", "No prediction possible."])

        with open(log_file, "a") as logs:
            logs.write("%s      %s                                                   %s"
                       % (time.asctime(time.localtime(time.time())), " ", "Gene orientation prediction finished! \n"))

    """-----------------------------------------------------------------------------------------------------------------
    ---------------------------------------------------GC-SKEW----------------------------------------------------------
    -----------------------------------------------------------------------------------------------------------------"""
    ORI_gc = ORI_ta = TER_gc = TER_ta = 0
    p_ori_gc = p_ori_ta = p_ter_gc = p_ter_ta = 1.0
    av_diff_gc = av_diff_ta = 0.0
    gc_cum = list()
    ta_cum = list()

    '''If -gc is specified: running GCscript.py'''
    if args.gc or args.all:
        with open(log_file, "a") as logs:
            logs.write("%s      %s" % (time.asctime(time.localtime(time.time())),
                                       "Starting GC-skew...\n"))

        '''If output file name is specified, enter this, otherwise use species name from file'''
        output_name = inner_path + "/" + species

        # Calls on GC and TA calculation
        ORI_gc, TER_gc, gc_cum, gc_normalized, perfect_gc, x_perfect_gc, \
        ORI_ta, TER_ta, ta_cum, ta_normalized, perfect_ta, x_perfect_ta, positions \
            = gc.main(record_dict, sequence, sequence_length, name_species, strain, output_name, args.graph)

        with open(log_file, "a") as logs:
            logs.write("%s      %s" % (time.asctime(time.localtime(time.time())),
                                       "...GC- and TA-skew calculations finished!\n"))

        # Calculates G, C, T and A content of the chromosome
        G_content, C_content, T_content, A_content \
            = gc.calculate_overall_gc(record_dict, positions, sequence.seq)

        # Calculates the average difference between the real and perfect data
        av_diff_gc, av_diff_ta = gc.diff_perf_real(perfect_gc, gc_normalized, perfect_ta, ta_normalized)

        # Finds all indices with value differing less than the average difference,
        # and the break points if found
        aroundori_gc, aroundter_gc, ori_break_gc, ter_break_gc = \
            gc.around_pos(gc_normalized, ORI_gc, TER_gc, av_diff_gc)
        aroundori_ta, aroundter_ta, ori_break_ta, ter_break_ta = \
            gc.around_pos(ta_normalized, ORI_ta, TER_ta, av_diff_ta)

        # Calculates slopes before and after ORI and TER
        av_gc_bf_ori, av_gc_af_ori, av_gc_bf_ter, av_gc_af_ter, av_ta_bf_ori, \
        av_ta_af_ori, av_ta_bf_ter, av_ta_af_ter, average_slope_gc, average_slope_ta, \
        av_ori_ter_gc, av_ori_ter_ta = gc.slopes(gc_normalized, ta_normalized, ORI_gc,
                                                 TER_gc, ORI_ta, TER_ta)

        # Initiates window size
        window_size = int(0.0001*sequence_length)

        ori_ter_gc = abs(gc_normalized[TER_gc] - gc_normalized[ORI_gc])
        ori_ter_ta = abs(ta_normalized[TER_ta] - ta_normalized[ORI_ta])

        '''If signficant dnaA boxes or dif motifs were found - uses this for evaluating other predictions.
        Otherwise, checks if significant rrna intervals - use this for evaluating predictions.'''
        if len(ori_dnaa_range) != 0 or len(ter_dif_range) != 0:
            if (ORI_gc in ori_dnaa_range or TER_gc in ter_dif_range) and \
                    (ORI_ta in ori_dnaa_range or TER_ta in ter_dif_range):
                if av_diff_gc <= av_diff_ta:
                    with open(log_file, "a") as logs:
                        logs.write("%s      %s                   " % (time.asctime(time.localtime(time.time())),
                                                                      "...Starting Bootstrap function...\n"))
                    # Starts bootstrap function
                    p_ori_gc, p_ter_gc \
                        = gc.gc_bootstrap(sequence_length, aroundter_gc,
                                          aroundori_gc, aroundter_ta, aroundori_ta, window_size, G_content, C_content,
                                          T_content, A_content, ori_break_gc, ter_break_gc, ori_break_ta, ter_break_ta,
                                          ORI_gc, TER_gc, log_file, ori_dnaa_range, ter_dif_range)
                    p_ori_ta = 1.0
                    p_ter_ta = 1.0
                    with open(log_file, "a") as logs:
                        logs.write("%s" % "   Done! \n")

                elif av_diff_ta < av_diff_gc:
                    with open(log_file, "a") as logs:
                        logs.write("%s      %s                   " % (time.asctime(time.localtime(time.time())),
                                                                      "...Starting Bootstrap function...\n"))
                    p_ori_ta, p_ter_ta \
                        = gc.gc_bootstrap(sequence_length, aroundter_gc,
                                          aroundori_gc, aroundter_ta, aroundori_ta, window_size, G_content, C_content,
                                          T_content, A_content, ori_break_gc, ter_break_gc, ori_break_ta, ter_break_ta,
                                          ORI_ta, TER_ta, log_file, ori_dnaa_range, ter_dif_range)
                    p_ori_gc = 1.0
                    p_ter_gc = 1.0
                    with open(log_file, "a") as logs:
                        logs.write("%s" % "   Done! \n")

            elif (ORI_gc in ori_rrna_range or TER_gc in ter_rrna_range) and \
                    (ORI_ta in ori_rrna_range or TER_ta in ter_rrna_range):
                if av_diff_gc <= av_diff_ta:
                    with open(log_file, "a") as logs:
                        logs.write("%s      %s                   " % (time.asctime(time.localtime(time.time())),
                                                                      "...Starting Bootstrap function...\n"))
                    # Starts bootstrap function
                    p_ori_gc, p_ter_gc \
                        = gc.gc_bootstrap(sequence_length, aroundter_gc,
                                          aroundori_gc, aroundter_ta, aroundori_ta, window_size, G_content, C_content,
                                          T_content, A_content, ori_break_gc, ter_break_gc, ori_break_ta, ter_break_ta,
                                          ORI_gc, TER_gc, log_file, ori_dnaa_range, ter_dif_range)
                    p_ori_ta = 1.0
                    p_ter_ta = 1.0
                    with open(log_file, "a") as logs:
                        logs.write("%s" % "   Done! \n")

                elif av_diff_ta < av_diff_gc:
                    with open(log_file, "a") as logs:
                        logs.write("%s      %s                   " % (time.asctime(time.localtime(time.time())),
                                                                      "...Starting Bootstrap function...\n"))
                    p_ori_ta, p_ter_ta \
                        = gc.gc_bootstrap(sequence_length, aroundter_gc,
                                          aroundori_gc, aroundter_ta, aroundori_ta, window_size, G_content, C_content,
                                          T_content, A_content, ori_break_gc, ter_break_gc, ori_break_ta, ter_break_ta,
                                          ORI_ta, TER_ta, log_file, ori_dnaa_range, ter_dif_range)
                    p_ori_gc = 1.0
                    p_ter_gc = 1.0
                    with open(log_file, "a") as logs:
                        logs.write("%s" % "   Done! \n")

            else:
                with open(log_file, "a") as logs:
                    logs.write("%s      %s" % (time.asctime(time.localtime(time.time())),
                                               "GC-skew predictions not within probable range. "
                                               "Bootstrap not performed.\n"))
                p_ori_gc = 1.0
                p_ter_gc = 1.0
                p_ori_ta = 1.0
                p_ter_ta = 1.0

        elif len(ori_rrna_range) != 0 or len(ter_rrna_range) != 0:
            if (ORI_gc in ori_rrna_range and TER_gc in ter_rrna_range) or \
                    (ORI_ta in ori_rrna_range and TER_ta in ter_rrna_range):
                if av_diff_gc <= av_diff_ta:
                    with open(log_file, "a") as logs:
                        logs.write("%s      %s                   " % (time.asctime(time.localtime(time.time())),
                                                                      "...Starting Bootstrap function...\n"))
                    # Starts bootstrap function
                    p_ori_gc, p_ter_gc \
                        = gc.gc_bootstrap(sequence_length, aroundter_gc,
                                          aroundori_gc, aroundter_ta, aroundori_ta, window_size, G_content, C_content,
                                          T_content, A_content, ori_break_gc, ter_break_gc, ori_break_ta, ter_break_ta,
                                          ORI_gc, TER_gc, log_file, ori_rrna_range, ter_rrna_range)
                    p_ori_ta = 1.0
                    p_ter_ta = 1.0
                    with open(log_file, "a") as logs:
                        logs.write("%s" % "   Done! \n")
                elif av_diff_ta < av_diff_gc:
                    with open(log_file, "a") as logs:
                        logs.write("%s      %s                   " % (time.asctime(time.localtime(time.time())),
                                                                      "...Starting Bootstrap function...\n"))
                    p_ori_ta, p_ter_ta \
                        = gc.gc_bootstrap(sequence_length, aroundter_gc,
                                          aroundori_gc, aroundter_ta, aroundori_ta, window_size, G_content, C_content,
                                          T_content, A_content, ori_break_gc, ter_break_gc, ori_break_ta, ter_break_ta,
                                          ORI_ta, TER_ta, log_file, ori_rrna_range, ter_rrna_range)
                    p_ori_gc = 1.0
                    p_ter_gc = 1.0
                    with open(log_file, "a") as logs:
                        logs.write("%s" % "   Done! \n")

                else:
                    with open(log_file, "a") as logs:
                        logs.write("%s      %s" % (time.asctime(time.localtime(time.time())),
                                                   "GC-skew predictions not within probable range. "
                                                   "Bootstrap not performed.\n"))
                    p_ori_gc = 1.0
                    p_ter_gc = 1.0
                    p_ori_ta = 1.0
                    p_ter_ta = 1.0

        else:
            if av_diff_gc <= av_diff_ta:
                with open(log_file, "a") as logs:
                    logs.write("%s      %s              " % (time.asctime(time.localtime(time.time())),
                                                                  "...Starting Bootstrap function...\n"))

                p_ori_gc, p_ter_gc = gc.gc_bootstrap(sequence_length, aroundter_gc, aroundori_gc,
                                                     aroundter_ta, aroundori_ta, window_size, G_content, C_content,
                                                     T_content, A_content, ori_break_gc, ter_break_gc, ori_break_ta,
                                                     ter_break_ta, ORI_gc, TER_gc, log_file,
                                                     ori_rrna_range, ter_rrna_range)
                p_ori_ta = 1.0
                p_ter_ta = 1.0

                with open(log_file, "a") as logs:
                    logs.write("%s" % "   Done! \n")
            elif av_diff_ta < av_diff_gc:
                with open(log_file, "a") as logs:
                    logs.write("%s      %s               " % (time.asctime(time.localtime(time.time())),
                                                                  "...Starting Bootstrap function...\n"))

                p_ori_ta, p_ter_ta = gc.gc_bootstrap(sequence_length, aroundter_gc, aroundori_gc,
                                                     aroundter_ta, aroundori_ta, window_size, G_content, C_content,
                                                     T_content, A_content, ori_break_gc, ter_break_gc, ori_break_ta,
                                                     ter_break_ta, ORI_ta, TER_ta, log_file,
                                                     ori_rrna_range, ter_rrna_range)
                p_ori_gc = 1.0
                p_ter_gc = 1.0

                with open(log_file, "a") as logs:
                    logs.write("%s" % "   Done! \n")
            else:
                with open(log_file, "a") as logs:
                    logs.write("%s      %s" % (time.asctime(time.localtime(time.time())),
                                               "GC-skew predictions not within probable range. "
                                               "Bootstrap not performed.\n"))
                p_ori_gc = 1.0
                p_ter_gc = 1.0
                p_ori_ta = 1.0
                p_ter_ta = 1.0

        # Save results in csv file
        comment_gc = ""
        comment_ta = ""
        comment = ""
        if p_ori_gc > alpha:
            comment_gc = "NB! p-value above significance level, \nORI prediction should not be trusted!"
        if p_ter_gc > alpha and comment == "":
            comment_gc = "NB! p-value above significance level, \nTER prediction should not be trusted!"
        elif p_ter_gc > alpha < p_ori_gc:
            comment_gc = "NB! Both p-values above signficance level, \nGC-skew should not be trusted!"

        if p_ori_ta > alpha:
            comment_ta = "NB! p-value above significance level, \nORI prediction should not be trusted!"
        if p_ter_ta > alpha and comment == "":
            comment_ta = "NB! p-value above significance level, \nTER prediction should not be trusted!"
        elif p_ori_ta > alpha < p_ter_ta:
            comment_ta = "NB! Both p-values above signficance level, \nTA-skew should not be trusted!"

        with open(csv_file, mode='a') as results:
            teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
            teritori_writer.writerow(["GC-skew", ORI_gc, p_ori_gc, TER_gc, p_ter_gc, comment_gc])
            teritori_writer.writerow(["TA-skew", ORI_ta, p_ori_ta, TER_ta, p_ter_ta, comment_ta])

        # Skew prediction finished
        with open(log_file, "a") as logs:
            logs.write("%s      %s                                                   %s"
                       % (time.asctime(time.localtime(time.time())), " ", "GC-skew prediction finished! \n"))

    """-----------------------------------------------------------------------------------------------------------------
    --------------------------------------------COMBINING THE RESULTS---------------------------------------------------
    -----------------------------------------------------------------------------------------------------------------"""
    with open(log_file, "a") as logs:
        logs.write("%s      %s" % (time.asctime(time.localtime(time.time())),
                                   "Calculating final position..."))

    '''Final prediction variables'''
    ORI_go_range = list()
    TER_go_range = list()

    # Calculates maximum expected difference between ORI and TER
    max_perc_genome = round(sequence_length * 0.65)
    min_perc_genome = int(sequence_length * 0.35)

    # Interval size
    intsize = 250

    # Add final prediction to file. 2 TERs if linear.
    if linear:
        with open(csv_file, mode='a') as results:
            teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
            teritori_writer.writerow([""])
            teritori_writer.writerow(["FINAL PREDICTION:"])
            teritori_writer.writerow(["", "ORI", "p-val\n(ori)", "TER 1", "TER 2", "p-val\n(ter)", "Comment"])
    else:
        with open(csv_file, mode='a') as results:
            teritori_writer = csv.writer(results, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
            teritori_writer.writerow([""])
            teritori_writer.writerow(["FINAL PREDICTION:"])
            teritori_writer.writerow(["", "ORI", "p-val\n(ori)", "TER", "p-val\n(ter)", "Comment"])

    fp.combine_all(ori_rrna_range, ori_dnaa_range, ori_go_range, dnaa_max_pos,
                   ORI_gc, ORI_ta, p_ori_gc, p_ori_ta, p_ori_go, ter_rrna_range,
                   ter_dif_range, ter_go_range, dif_max_pos, TER_gc, TER_ta,
                   p_ter_gc, p_ter_ta, p_ter_go, alpha, csv_file, sequence_length,
                   av_diff_gc, av_diff_ta, linear, args.cm, args.gc, args.all, args.go)

    # Final prediction finished
    with open(log_file, "a") as logs:
        logs.write("%s" % "Done! \n")

    """-----------------------------------------------------------------------------------------------------------------
    ----------------------------------------------------PLOT------------------------------------------------------------
    -----------------------------------------------------------------------------------------------------------------"""
    fig_title = name_species
    if args.graph:
        with open(log_file, "a") as logs:
            logs.write("%s      %s" % (time.asctime(time.localtime(time.time())),
                                       "Creating plot..."))
        if args.output:
            title = args.output

        final_result = list()
        with open(csv_file, 'r') as csvFile:
            reader = csv.reader(csvFile)
            for row in reader:
                final_result.append(row)
        csvFile.close()

        fig = plt.figure()
        sub = fig.add_subplot(111)

        '''If all is chose, motifs are visualized'''
        if args.all or (args.gc and args.cm):
            for i in dnaa_sorted:
                if dnaa_sorted[i][0] > dnaa_sorted[i][1]:
                    x = list(range(dnaa_sorted[i][1], sequence_length)) + \
                        list(range(0, dnaa_sorted[i][0]))
                else:
                    x = list(range(dnaa_sorted[i][0], dnaa_sorted[i][1]))
                y = [0 for x in range(len(x))]
                if dnaa_sorted[i][2] == "+":
                    sub.plot(x, y, 'b>')
                elif dnaa_sorted[i][2] == "-":
                    sub.plot(x, y, 'b<')

            for i in dif_sorted:
                if dif_sorted[i][0] > dif_sorted[i][1]:
                    x = list(range(dif_sorted[i][1], sequence_length)) + \
                        list(range(0, dif_sorted[i][0]))
                else:
                    x = list(range(dif_sorted[i][0], dif_sorted[i][1]))
                y = [1 for x in range(len(x))]
                if dif_sorted[i][2] == "+":
                    sub.plot(x, y, 'r>')
                elif dif_sorted[i][2] == "-":
                    sub.plot(x, y, 'r<')

            for i in rrnas:
                x = list(range(i[0], i[1]))
                y = [0.5 for x in range(len(x))]
                if i[2] == 1:
                    sub.plot(x, y, 'm>')
                elif i[2] == -1:
                    sub.plot(x, y, 'm<')

        '''If go is chosen, go is plotted.'''
        if args.go and not args.all and not args.cm and not args.gc:
            x_go = np.arange(len_data_plot)
            sub.plot(x_go, normalized_go, color='black', linewidth=2)

        '''If gc skew or all is chosen'''
        if args.gc or args.all:
            x = x_perfect_gc
            sub.plot(x, gc_normalized, color='black', linewidth=2)
            sub.plot(x, ta_normalized, color='darkblue', linewidth=2)

        plt.suptitle(fig_title)

        if args.all or args.gc or args.cm:
            plt.xlabel('Sequence position')
        elif args.go and not args.all and not args.gc and not args.cm:
            plt.xlabel('Gene number')

        if args.all or (args.gc and args.cm and args.go) or (args.gc and args.cm):
            black_patch = mpatches.Patch(color='black', label='GC-skew')
            dark_patch = mpatches.Patch(color='darkblue', label='TA-skew')
            blue_patch = mpatches.Patch(color='blue', label='dnA boxes')
            red_patch = mpatches.Patch(color='red', label='dif motifs')
            magenta_patch = mpatches.Patch(color='magenta', label='rRNAs')
            sub.legend(handles=[black_patch, dark_patch, blue_patch, red_patch, magenta_patch], loc='upper center',
                       bbox_to_anchor=(0.5, -0.05), shadow=True, ncol=5)
        elif args.go and not args.gc and not args.all and not args.cm:
            if best_ori == "cum-":
                labletext = "Gene orientation (cumulative)"
            elif best_ori == "sumdiff+":
                labletext = "Gene orientation (sumdiff)"

            black_patch = mpatches.Patch(color='black', label='Gene-orientation')
            sub.legend(handles=[black_patch], loc='upper center',
                       bbox_to_anchor=(0.5, -0.05), shadow=True, ncol=1)
        elif args.gc and not args.go and not args.all and not args.cm:
            black_patch = mpatches.Patch(color='black', label='GC-skew')
            dark_patch = mpatches.Patch(color='darkblue', label='TA-skew')
            sub.legend(handles=[black_patch, dark_patch], loc='upper center',
                       bbox_to_anchor=(0.5, -0.05), shadow=True, ncol=3)
        elif args.gc and args.go:
            black_patch = mpatches.Patch(color='black', label='GC-skew')
            dark_patch = mpatches.Patch(color='darkblue', label='TA-skew')
            sub.legend(handles=[black_patch, dark_patch], loc='upper center',
                       bbox_to_anchor=(0.5, -0.05), shadow=True, ncol=2)
        TER = TER_1 = TER_2 = 0
        p_ori = p_ter = 0.0

        ORI = final_result[-1][1]
        if ORI.isdigit():
            ORI = int(ORI)
            if final_result[-1][2] == "":
                p_ori = '---'
            else:
                p_ori = str(final_result[-1][2])

            sub.axvline(ORI, linestyle='--', color='grey')
        else:
            p_ori = '---'

        if not args.cm:
            if linear:
                len(final_result)
                TER_1 = final_result[-1][3]
                TER_2 = final_result[-1][4]
                print(TER_1, TER_2)
                if TER_1.isdigit() and TER_2.isdigit():
                    TER_1 = int(TER_1)
                    TER_2 = int(TER_2)

                    if final_result[-1][5] == "":
                        p_ter = '---'
                    else:
                        p_ter = str(final_result[-1][5])

                    sub.axvline(TER_1, linestyle='--', color='grey')
                    sub.axvline(TER_2, linestyle='--', color='grey')
                else:
                    p_ter = '---'

                sub.annotate('ORI:' + str(ORI) + ', p-val: ' + p_ori +
                             '          TER:' + str(TER_1) + ' & ' + str(TER_2) + ', p-val: ' + p_ter,
                             xy=(0.07, 1.02), xycoords='axes fraction')
            elif not linear:
                TER = final_result[-1][3]
                if TER.isdigit():
                    TER = int(TER)
                    if final_result[-1][4] == "":
                        p_ter = '---'
                    else:
                        p_ter = str(final_result[-1][4])

                    sub.axvline(TER, linestyle='--', color='grey')
                else:
                    p_ter = '---'

                sub.annotate('ORI:' + str(ORI) + ', p-val: ' + p_ori + '        TER:' + str(TER) + ', p-val: ' + p_ter,
                             xy=(0.1, 1.02), xycoords='axes fraction')

        if args.all or args.gc or args.go or (args.cm and args.go and args.gc) or (args.go and args.gc) \
            or (args.gc and args.cm):
            if args.output:
                plt.savefig(inner_path + "/" + args.output + ".png", format="png")
            else:
                plt.savefig(inner_path + "/" + species + ".png", format="png")

    with open(log_file, "a") as logs:
        logs.write("%s" % "Done! \n\n")

    '''-----------------------------------------terITori finished----------------------------------------------------'''

    with open(log_file, "a") as logs:
        logs.write("%s      %s" % (time.asctime(time.localtime(time.time())),
                                   "terITori successful!"))


if __name__ == "__main__":
    main()